"""Focused tests for beta lab route guards and resolver contract."""

from __future__ import annotations

import asyncio
import os
import secrets
import sys

import pytest
from fastapi import HTTPException
from fastapi.testclient import TestClient

from bloom_lims.api.v1.beta_lab import require_external_ursa_read
from bloom_lims.api.v1.dependencies import (
    APIUser,
    require_external_token_auth,
    require_external_ursa_api_enabled,
)
from bloom_lims.auth.rbac import ENABLE_ATLAS_API_GROUP, ENABLE_URSA_API_GROUP
from bloom_lims.bobjs import BloomObj
from bloom_lims.db import get_parent_lineages

os.environ["BLOOM_DEV_AUTH_BYPASS"] = "true"
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from main import app  # noqa: E402


def _token_user(groups: list[str]) -> APIUser:
    return APIUser(
        email="token-user@example.com",
        user_id="token-user",
        roles=["INTERNAL_READ_ONLY"],
        groups=groups,
        auth_source="token",
    )


def test_require_external_ursa_api_enabled_rejects_missing_group():
    with pytest.raises(HTTPException) as exc:
        asyncio.run(require_external_ursa_api_enabled(_token_user(groups=[])))
    assert exc.value.status_code == 403
    assert "ENABLE_URSA_API" in str(exc.value.detail)


def test_require_external_ursa_api_enabled_allows_group_member():
    user = _token_user(groups=[ENABLE_URSA_API_GROUP])
    resolved = asyncio.run(require_external_ursa_api_enabled(user))
    assert resolved.user_id == "token-user"


def test_run_resolver_requires_full_key_query_params():
    client = TestClient(app)
    app.dependency_overrides[require_external_ursa_read] = lambda: APIUser(
        email="resolver-user@example.com",
        user_id="resolver-user",
        roles=["INTERNAL_READ_ONLY"],
        groups=[ENABLE_URSA_API_GROUP],
        auth_source="token",
    )
    try:
        response = client.get("/api/v1/external/atlas/beta/runs/RUN-1/resolve")
        assert response.status_code == 422
    finally:
        app.dependency_overrides.pop(require_external_ursa_read, None)


def test_material_registration_links_fulfillment_items_on_container_and_patient_on_specimen(bdb):
    def _atlas_rw_user() -> APIUser:
        token = secrets.token_hex(8)
        return APIUser(
            email="atlas-beta@example.com",
            user_id=f"atlas-user-{token}",
            roles=["INTERNAL_READ_WRITE"],
            groups=[ENABLE_ATLAS_API_GROUP],
            auth_source="token",
            is_service_account=True,
            token_scope="internal_rw",
            token_id=f"token-{token}",
        )

    def _atlas_refs(instance):
        refs: list[dict] = []
        for lineage in get_parent_lineages(instance):
            if lineage.is_deleted or lineage.relationship_type != "has_external_reference":
                continue
            child = lineage.child_instance
            if child is None or child.is_deleted:
                continue
            payload = child.json_addl or {}
            props = payload.get("properties") if isinstance(payload, dict) else {}
            if not isinstance(props, dict):
                continue
            if str(props.get("provider") or "").strip() != "atlas":
                continue
            refs.append(props)
        return refs

    app.dependency_overrides[require_external_token_auth] = _atlas_rw_user
    client = TestClient(app)
    try:
        atlas_context = {
            "atlas_tenant_id": f"tenant-{secrets.token_hex(8)}",
            "atlas_trf_euid": f"trf-{secrets.token_hex(8)}",
            "atlas_patient_euid": f"patient-{secrets.token_hex(8)}",
            "atlas_testkit_euid": f"kit-{secrets.token_hex(8)}",
            "atlas_shipment_euid": f"shipment-{secrets.token_hex(8)}",
            "atlas_organization_site_euid": f"site-{secrets.token_hex(8)}",
            "fulfillment_items": [
                {
                    "atlas_test_euid": f"test-{secrets.token_hex(8)}",
                    "atlas_test_fulfillment_item_euid": f"proc-{secrets.token_hex(8)}",
                }
            ],
        }
        material = client.post(
            "/api/v1/external/atlas/beta/materials",
            headers={"Idempotency-Key": f"idem-material-{secrets.token_hex(8)}"},
            json={"specimen_name": "container-first-link-check", "atlas_context": atlas_context},
        )
        assert material.status_code == 200, material.text
        material_body = material.json()
        specimen_euid = material_body["specimen_euid"]
        container_euid = material_body["container_euid"]
        assert material_body["atlas_context"]["atlas_testkit_euid"] == atlas_context["atlas_testkit_euid"]
        assert material_body["atlas_context"]["atlas_shipment_euid"] == atlas_context["atlas_shipment_euid"]
        assert (
            material_body["atlas_context"]["atlas_organization_site_euid"]
            == atlas_context["atlas_organization_site_euid"]
        )

        bobj = BloomObj(bdb)
        specimen = bobj.get_by_euid(specimen_euid)
        container = bobj.get_by_euid(container_euid)
        assert container.created_dt <= specimen.created_dt

        specimen_refs = _atlas_refs(specimen)
        container_refs = _atlas_refs(container)
        assert any(
            str(ref.get("reference_type")) == "atlas_patient"
            and str(ref.get("atlas_patient_euid")) == atlas_context["atlas_patient_euid"]
            for ref in specimen_refs
        )
        assert not any(
            str(ref.get("reference_type")) == "atlas_test_process_item"
            for ref in specimen_refs
        )
        assert any(
            str(ref.get("reference_type")) == "atlas_test_process_item"
            and str(ref.get("atlas_test_fulfillment_item_euid"))
            == atlas_context["fulfillment_items"][0]["atlas_test_fulfillment_item_euid"]
            for ref in container_refs
        )
        assert any(
            str(ref.get("reference_type")) == "atlas_trf"
            and str(ref.get("atlas_trf_euid")) == atlas_context["atlas_trf_euid"]
            for ref in container_refs
        )
        assert any(
            str(ref.get("reference_type")) == "atlas_testkit"
            and str(ref.get("atlas_testkit_euid")) == atlas_context["atlas_testkit_euid"]
            for ref in container_refs
        )
        assert any(
            str(ref.get("reference_type")) == "atlas_shipment"
            and str(ref.get("atlas_shipment_euid")) == atlas_context["atlas_shipment_euid"]
            for ref in container_refs
        )
        assert any(
            str(ref.get("reference_type")) == "atlas_organization_site"
            and str(ref.get("atlas_organization_site_euid"))
            == atlas_context["atlas_organization_site_euid"]
            for ref in container_refs
        )

        queued = client.post(
            f"/api/v1/external/atlas/beta/queues/extraction_prod/items/{container_euid}",
            headers={"Idempotency-Key": f"idem-queue-{secrets.token_hex(8)}"},
            json={"metadata": {"source": "test_beta_lab"}},
        )
        assert queued.status_code == 200, queued.text
        extraction = client.post(
            "/api/v1/external/atlas/beta/extractions",
            headers={"Idempotency-Key": f"idem-extract-{secrets.token_hex(8)}"},
            json={
                "source_specimen_euid": specimen_euid,
                "well_name": "A1",
                "extraction_type": "gdna",
                "atlas_test_fulfillment_item_euid": atlas_context["fulfillment_items"][0][
                    "atlas_test_fulfillment_item_euid"
                ],
            },
        )
        assert extraction.status_code == 200, extraction.text
    finally:
        client.close()
        app.dependency_overrides.pop(require_external_token_auth, None)


def test_empty_tube_create_and_specimen_update_use_collection_event_reference(bdb):
    def _atlas_rw_user() -> APIUser:
        token = secrets.token_hex(8)
        return APIUser(
            email="atlas-beta@example.com",
            user_id=f"atlas-user-{token}",
            roles=["INTERNAL_READ_WRITE"],
            groups=[ENABLE_ATLAS_API_GROUP],
            auth_source="token",
            is_service_account=True,
            token_scope="internal_rw",
            token_id=f"token-{token}",
        )

    def _atlas_refs(instance):
        refs: list[dict] = []
        for lineage in get_parent_lineages(instance):
            if lineage.is_deleted or lineage.relationship_type != "has_external_reference":
                continue
            child = lineage.child_instance
            if child is None or child.is_deleted:
                continue
            payload = child.json_addl or {}
            props = payload.get("properties") if isinstance(payload, dict) else {}
            if not isinstance(props, dict):
                continue
            if str(props.get("provider") or "").strip() != "atlas":
                continue
            refs.append(props)
        return refs

    app.dependency_overrides[require_external_token_auth] = _atlas_rw_user
    client = TestClient(app)
    try:
        tenant_id = f"tenant-{secrets.token_hex(8)}"
        shipment_euid = f"shipment-{secrets.token_hex(8)}"
        testkit_euid = f"kit-{secrets.token_hex(8)}"
        site_euid = f"site-{secrets.token_hex(8)}"
        create_tube = client.post(
            "/api/v1/external/atlas/beta/tubes",
            headers={"Idempotency-Key": f"idem-tube-{secrets.token_hex(8)}"},
            json={
                "atlas_context": {
                    "atlas_tenant_id": tenant_id,
                    "atlas_shipment_euid": shipment_euid,
                    "atlas_testkit_euid": testkit_euid,
                    "atlas_organization_site_euid": site_euid,
                }
            },
        )
        assert create_tube.status_code == 200, create_tube.text
        tube_body = create_tube.json()
        container_euid = tube_body["container_euid"]

        collection_event_euid = f"cev-{secrets.token_hex(8)}"
        specimen_resp = client.post(
            "/api/v1/external/atlas/beta/materials",
            headers={"Idempotency-Key": f"idem-specimen-{secrets.token_hex(8)}"},
            json={
                "container_euid": container_euid,
                "atlas_context": {
                    "atlas_tenant_id": tenant_id,
                    "atlas_shipment_euid": shipment_euid,
                    "atlas_testkit_euid": testkit_euid,
                    "atlas_organization_site_euid": site_euid,
                    "atlas_collection_event_euid": collection_event_euid,
                    "collection_event_snapshot": {
                        "collection_event_euid": collection_event_euid,
                        "collection_type": "venipuncture",
                    },
                },
            },
        )
        assert specimen_resp.status_code == 200, specimen_resp.text
        specimen_body = specimen_resp.json()

        bobj = BloomObj(bdb)
        container = bobj.get_by_euid(container_euid)
        specimen = bobj.get_by_euid(specimen_body["specimen_euid"])

        container_refs = _atlas_refs(container)
        specimen_refs = _atlas_refs(specimen)
        assert any(
            str(ref.get("reference_type")) == "atlas_shipment"
            and str(ref.get("atlas_shipment_euid")) == shipment_euid
            for ref in container_refs
        )
        assert any(
            str(ref.get("reference_type")) == "atlas_testkit"
            and str(ref.get("atlas_testkit_euid")) == testkit_euid
            for ref in container_refs
        )
        assert any(
            str(ref.get("reference_type")) == "atlas_collection_event"
            and str(ref.get("atlas_collection_event_euid")) == collection_event_euid
            and str(ref.get("collection_event_snapshot", {}).get("collection_type")) == "venipuncture"
            for ref in specimen_refs
        )
        assert not any(str(ref.get("reference_type")) == "atlas_patient" for ref in specimen_refs)

        patched = client.patch(
            f"/api/v1/external/atlas/beta/specimens/{specimen_body['specimen_euid']}",
            json={
                "atlas_context": {
                    "atlas_tenant_id": tenant_id,
                    "atlas_collection_event_euid": collection_event_euid,
                    "collection_event_snapshot": {
                        "collection_event_euid": collection_event_euid,
                        "collection_type": "fingerstick",
                        "expected_name": "Pat Ient",
                    },
                }
            },
        )
        assert patched.status_code == 200, patched.text

        updated_specimen_refs = _atlas_refs(bobj.get_by_euid(specimen_body["specimen_euid"]))
        assert any(
            str(ref.get("reference_type")) == "atlas_collection_event"
            and str(ref.get("collection_event_snapshot", {}).get("collection_type")) == "fingerstick"
            and str(ref.get("collection_event_snapshot", {}).get("expected_name")) == "Pat Ient"
            for ref in updated_specimen_refs
        )
        assert not any(
            str(ref.get("reference_type")) == "atlas_patient" for ref in updated_specimen_refs
        )
    finally:
        client.close()
        app.dependency_overrides.pop(require_external_token_auth, None)



"""Focused tests for Bloom RBAC/token auth additions."""

import asyncio
import os
import sys
from datetime import datetime, timedelta
from time import time_ns

import pytest
from fastapi import HTTPException
from fastapi.testclient import TestClient

from bloom_lims.api.v1.dependencies import APIUser, require_external_token_auth, require_write

# Ensure auth bypass is active for API-client tests.
os.environ["BLOOM_DEV_AUTH_BYPASS"] = "true"
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from main import app  # noqa: E402


@pytest.fixture
def client():
    return TestClient(app)


def _create_token(client: TestClient) -> str:
    token_name = f"pytest-token-{time_ns()}"
    create_response = client.post(
        "/api/v1/user-tokens",
        json={
            "token_name": token_name,
            "scope": "internal_ro",
            "expires_in_days": 30,
        },
    )
    assert create_response.status_code == 200
    payload = create_response.json()
    assert payload["token"]["token_name"] == token_name
    assert payload["plaintext_token"].startswith("blm_")
    return payload["token"]["token_id"]


def test_require_write_blocks_read_only_user():
    user = APIUser(
        email="ro@example.com",
        user_id="ro-user-1",
        roles=["INTERNAL_READ_ONLY"],
    )
    with pytest.raises(HTTPException) as exc:
        asyncio.run(require_write(user))
    assert exc.value.status_code == 403
    assert "Write permission required" in exc.value.detail


def test_require_external_token_auth_blocks_non_token_user():
    user = APIUser(
        email="session@example.com",
        user_id="session-user-1",
        roles=["INTERNAL_READ_WRITE"],
        auth_source="session",
    )
    with pytest.raises(HTTPException) as exc:
        asyncio.run(require_external_token_auth(user))
    assert exc.value.status_code == 401


def test_auth_me_includes_rbac_fields(client):
    response = client.get("/api/v1/auth/me")
    assert response.status_code == 200
    payload = response.json()
    assert "roles" in payload
    assert "groups" in payload
    assert "permissions" in payload
    assert "auth_source" in payload


def test_api_root_lists_new_auth_endpoints(client):
    response = client.get("/api/v1/")
    assert response.status_code == 200
    endpoints = response.json()["endpoints"]
    assert "user_tokens" in endpoints
    assert "admin_auth" in endpoints


def test_external_specimen_endpoints_require_external_token_auth(client):
    response = client.get("/api/v1/external/specimens/by-reference?order_number=ORD-1")
    assert response.status_code == 401
    assert "requires Bloom API bearer token" in response.json()["detail"]


def test_user_tokens_endpoints_create_list_usage_revoke(client):
    token_id = _create_token(client)

    list_response = client.get("/api/v1/user-tokens")
    assert list_response.status_code == 200
    list_payload = list_response.json()
    assert any(item["token_id"] == token_id for item in list_payload["items"])

    usage_response = client.get(f"/api/v1/user-tokens/{token_id}/usage")
    assert usage_response.status_code == 200
    usage_payload = usage_response.json()
    assert "items" in usage_payload
    assert "total" in usage_payload

    revoke_response = client.delete(f"/api/v1/user-tokens/{token_id}")
    assert revoke_response.status_code == 200
    revoked = revoke_response.json()
    assert revoked["token"]["status"] == "REVOKED"


def test_user_tokens_default_expiry_is_48_hours(client):
    response = client.post(
        "/api/v1/user-tokens",
        json={
            "token_name": f"pytest-default-expiry-{time_ns()}",
            "scope": "internal_ro",
        },
    )
    assert response.status_code == 200
    payload = response.json()

    created_at = datetime.fromisoformat(payload["token"]["created_at"])
    expires_at = datetime.fromisoformat(payload["token"]["expires_at"])
    lifetime = expires_at - created_at
    assert timedelta(hours=47, minutes=55) <= lifetime <= timedelta(hours=48, minutes=5)


def test_admin_group_membership_endpoints_list_add_get_delete(client):
    list_groups_response = client.get("/api/v1/admin/groups")
    assert list_groups_response.status_code == 200
    groups_payload = list_groups_response.json()
    assert groups_payload["total"] >= 1

    target_group = "API_ACCESS"
    member_user_id = f"member-user-{time_ns()}"

    add_response = client.post(
        f"/api/v1/admin/groups/{target_group}/members",
        json={"user_id": member_user_id},
    )
    assert add_response.status_code == 200
    add_payload = add_response.json()
    assert add_payload["group_code"] == target_group
    assert add_payload["user_id"] == member_user_id

    list_members_response = client.get(f"/api/v1/admin/groups/{target_group}/members")
    assert list_members_response.status_code == 200
    members_payload = list_members_response.json()
    assert any(item["user_id"] == member_user_id for item in members_payload["items"])

    remove_response = client.delete(f"/api/v1/admin/groups/{target_group}/members/{member_user_id}")
    assert remove_response.status_code == 200
    remove_payload = remove_response.json()
    assert remove_payload["group_code"] == target_group
    assert remove_payload["user_id"] == member_user_id


def test_admin_user_token_endpoints_list_usage_delete(client):
    token_id = _create_token(client)

    list_response = client.get("/api/v1/admin/user-tokens")
    assert list_response.status_code == 200
    list_payload = list_response.json()
    assert any(item["token_id"] == token_id for item in list_payload["items"])

    usage_response = client.get(f"/api/v1/admin/user-tokens/{token_id}/usage")
    assert usage_response.status_code == 200
    usage_payload = usage_response.json()
    assert "items" in usage_payload
    assert "total" in usage_payload

    revoke_response = client.delete(f"/api/v1/admin/user-tokens/{token_id}")
    assert revoke_response.status_code == 200
    revoke_payload = revoke_response.json()
    assert revoke_payload["token_id"] == token_id
    assert revoke_payload["status"] == "REVOKED"



"""
BLOOM LIMS API v1 - Subjects Endpoints

Endpoints for subject management (patients, donors, etc.).
"""

import logging
from typing import Any, Dict, List, Optional

from fastapi import APIRouter, Depends, HTTPException, Query
from pydantic import BaseModel, Field

from bloom_lims.db import get_parent_lineages
from .dependencies import require_api_auth, APIUser


logger = logging.getLogger(__name__)


router = APIRouter(prefix="/subjects", tags=["Subjects"])


def get_bdb(username: str = "api-user"):
    """Get database connection."""
    from bloom_lims.db import BLOOMdb3
    return BLOOMdb3(app_username=username)


class SubjectCreateSchema(BaseModel):
    """Schema for creating a subject."""
    template_euid: str = Field(..., description="Template EUID to create from")
    name: Optional[str] = Field(None, description="Subject name/identifier")
    external_id: Optional[str] = Field(None, description="External identifier")
    json_addl: Optional[Dict[str, Any]] = Field(None, description="Additional properties")


class SubjectUpdateSchema(BaseModel):
    """Schema for updating a subject."""
    name: Optional[str] = Field(None, description="Subject name/identifier")
    status: Optional[str] = Field(None, description="Subject status")
    json_addl: Optional[Dict[str, Any]] = Field(None, description="Additional properties")


@router.get("/", response_model=Dict[str, Any])
async def list_subjects(
    subtype: Optional[str] = Query(None, description="Filter by subtype"),
    status: Optional[str] = Query(None, description="Filter by status"),
    page: int = Query(1, ge=1),
    page_size: int = Query(50, ge=1, le=1000),
    user: APIUser = Depends(require_api_auth),
):
    """List subjects with optional filters."""
    try:
        bdb = get_bdb(user.email)

        query = bdb.session.query(bdb.Base.classes.generic_instance)
        query = query.filter(bdb.Base.classes.generic_instance.category == "subject")

        if subtype:
            query = query.filter(bdb.Base.classes.generic_instance.subtype == subtype.lower())
        if status:
            query = query.filter(bdb.Base.classes.generic_instance.bstatus == status)
        
        query = query.filter(bdb.Base.classes.generic_instance.is_deleted == False)
        
        total = query.count()
        offset = (page - 1) * page_size
        items = query.limit(page_size).offset(offset).all()
        
        return {
            "items": [
                {
                    "euid": obj.euid,
                    "name": obj.name,
                    "type": obj.type,
                    "subtype": obj.subtype,
                    "status": obj.bstatus,
                }
                for obj in items
            ],
            "total": total,
            "page": page,
            "page_size": page_size,
        }
    except Exception as e:
        logger.error(f"Error listing subjects: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.get("/{euid}")
async def get_subject(euid: str, user: APIUser = Depends(require_api_auth)):
    """Get a subject by EUID."""
    try:
        bdb = get_bdb(user.email)
        
        subject = bdb.session.query(bdb.Base.classes.generic_instance).filter(
            bdb.Base.classes.generic_instance.euid == euid,
            bdb.Base.classes.generic_instance.category == "subject",
        ).first()

        if not subject:
            raise HTTPException(status_code=404, detail=f"Subject not found: {euid}")

        return {
            "euid": subject.euid,
            "name": subject.name,
            "type": subject.type,
            "subtype": subject.subtype,
            "status": subject.bstatus,
            "json_addl": subject.json_addl,
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error getting subject {euid}: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.post("/", response_model=Dict[str, Any])
async def create_subject(
    data: SubjectCreateSchema,
    user: APIUser = Depends(require_api_auth),
):
    """Create a new subject from a template."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomObj
        from sqlalchemy.orm.attributes import flag_modified
        
        bo = BloomObj(bdb)
        bo.set_actor_context(user_id=user.user_id, email=user.email)
        result = bo.create_instances(data.template_euid)
        
        if not result or not result[0]:
            raise HTTPException(status_code=400, detail="Failed to create subject from template")
        
        subject = result[0][0]
        
        if data.name:
            subject.name = data.name
        if data.external_id:
            subject.json_addl = subject.json_addl or {}
            subject.json_addl.setdefault("properties", {})["external_id"] = data.external_id
            flag_modified(subject, "json_addl")
        if data.json_addl:
            existing = subject.json_addl or {}
            existing.update(data.json_addl)
            subject.json_addl = existing
            flag_modified(subject, "json_addl")
        
        bdb.session.commit()
        bo.track_user_interaction(
            subject.euid,
            relationship_type="user_created",
            user_id=user.user_id,
            email=user.email,
        )
        
        return {
            "success": True,
            "euid": subject.euid,
            "message": "Subject created successfully",
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error creating subject: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.put("/{euid}", response_model=Dict[str, Any])
async def update_subject(
    euid: str,
    data: SubjectUpdateSchema,
    user: APIUser = Depends(require_api_auth),
):
    """Update a subject."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomObj
        from sqlalchemy.orm.attributes import flag_modified
        bo = BloomObj(bdb)
        bo.set_actor_context(user_id=user.user_id, email=user.email)

        subject = bdb.session.query(bdb.Base.classes.generic_instance).filter(
            bdb.Base.classes.generic_instance.euid == euid,
            bdb.Base.classes.generic_instance.category == "subject",
        ).first()

        if not subject:
            raise HTTPException(status_code=404, detail=f"Subject not found: {euid}")

        if data.name is not None:
            subject.name = data.name
        if data.status is not None:
            subject.bstatus = data.status
        if data.json_addl is not None:
            existing = subject.json_addl or {}
            existing.update(data.json_addl)
            subject.json_addl = existing
            flag_modified(subject, "json_addl")

        bdb.session.commit()
        bo.track_user_interaction(
            subject.euid,
            relationship_type="user_updated",
            user_id=user.user_id,
            email=user.email,
        )

        return {
            "success": True,
            "euid": subject.euid,
            "message": "Subject updated successfully",
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error updating subject {euid}: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.delete("/{euid}", response_model=Dict[str, Any])
async def delete_subject(
    euid: str,
    hard_delete: bool = Query(False, description="Permanently delete"),
    user: APIUser = Depends(require_api_auth),
):
    """Delete a subject (soft delete by default)."""
    try:
        bdb = get_bdb(user.email)

        subject = bdb.session.query(bdb.Base.classes.generic_instance).filter(
            bdb.Base.classes.generic_instance.euid == euid,
            bdb.Base.classes.generic_instance.category == "subject",
        ).first()

        if not subject:
            raise HTTPException(status_code=404, detail=f"Subject not found: {euid}")

        if hard_delete:
            bdb.session.delete(subject)
        else:
            subject.is_deleted = True

        bdb.session.commit()

        return {
            "success": True,
            "euid": euid,
            "message": f"Subject {'permanently deleted' if hard_delete else 'soft deleted'}",
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error deleting subject {euid}: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.get("/{euid}/specimens", response_model=Dict[str, Any])
async def get_subject_specimens(euid: str, user: APIUser = Depends(require_api_auth)):
    """Get all specimens associated with a subject."""
    try:
        bdb = get_bdb(user.email)

        subject = bdb.session.query(bdb.Base.classes.generic_instance).filter(
            bdb.Base.classes.generic_instance.euid == euid,
            bdb.Base.classes.generic_instance.category == "subject",
        ).first()

        if not subject:
            raise HTTPException(status_code=404, detail=f"Subject not found: {euid}")

        specimens = []
        for lineage in get_parent_lineages(subject):
            if lineage.is_deleted:
                continue
            child = lineage.child_instance
            if child.category == "content" and child.type in ["specimen", "sample"]:
                specimens.append({
                    "euid": child.euid,
                    "name": child.name,
                    "type": child.type,
                    "subtype": child.subtype,
                    "status": child.bstatus,
                })

        return {
            "subject_euid": euid,
            "specimens": specimens,
            "count": len(specimens),
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error getting subject specimens {euid}: {e}")
        raise HTTPException(status_code=500, detail=str(e))



"""TapDB action recording for execution queue mutations."""

from __future__ import annotations

from datetime import UTC, datetime
from typing import Any

from daylily_tapdb.sequences import ensure_instance_prefix_sequence
from sqlalchemy import func, select
from sqlalchemy.orm import Session

from bloom_lims.domain.beta_actions import BETA_ACTION_TEMPLATE_PREFIX
from bloom_lims.tapdb_adapter import action_instance, action_instance_lineage, generic_template

EXECUTION_ACTION_GROUP = "execution_queue"
EXECUTION_ACTION_TEMPLATE_PREFIX = BETA_ACTION_TEMPLATE_PREFIX

_ACTION_TEMPLATE_DEFINITIONS: dict[str, dict[str, str]] = {
    "register_worker": {
        "name": "Register Worker",
        "description": "Registers or updates a persistent execution worker.",
    },
    "heartbeat_worker": {
        "name": "Heartbeat Worker",
        "description": "Updates worker heartbeat and liveness metadata.",
    },
    "claim_queue_item": {
        "name": "Claim Queue Item",
        "description": "Creates a lease and STARTED execution record for a queued subject.",
    },
    "renew_queue_lease": {
        "name": "Renew Queue Lease",
        "description": "Renews an active queue lease.",
    },
    "release_queue_lease": {
        "name": "Release Queue Lease",
        "description": "Releases an active queue lease without completion.",
    },
    "complete_queue_execution": {
        "name": "Complete Queue Execution",
        "description": "Completes queued execution and advances subject state.",
    },
    "fail_queue_execution": {
        "name": "Fail Queue Execution",
        "description": "Marks a queued execution attempt as retryable or terminal failure.",
    },
    "place_execution_hold": {
        "name": "Place Execution Hold",
        "description": "Places an explicit execution hold on a subject.",
    },
    "release_execution_hold": {
        "name": "Release Execution Hold",
        "description": "Releases an explicit execution hold.",
    },
    "requeue_subject": {
        "name": "Requeue Subject",
        "description": "Moves a subject into READY state for a queue/action pair.",
    },
    "cancel_subject_execution": {
        "name": "Cancel Subject Execution",
        "description": "Cancels execution for a subject.",
    },
    "expire_queue_lease": {
        "name": "Expire Queue Lease",
        "description": "Expires a stale queue lease for observability cleanup.",
    },
}


def _parse_template_code(template_code: str) -> tuple[str, str, str, str]:
    parts = [part for part in str(template_code or "").strip("/").split("/") if part]
    if len(parts) != 4:
        raise ValueError(f"Invalid template code: {template_code!r}")
    return parts[0], parts[1], parts[2], parts[3]


class ExecutionQueueActionRecorder:
    """Persists first-class action records for execution queue mutations."""

    def __init__(self, session: Session):
        self.session = session

    def find_replay(
        self,
        *,
        action_key: str,
        subject_euid: str | None,
        idempotency_key: str,
    ) -> action_instance | None:
        if action_key not in _ACTION_TEMPLATE_DEFINITIONS:
            raise ValueError(f"Unsupported execution action key: {action_key}")
        clean_subject_euid = str(subject_euid or "").strip()
        clean_idempotency_key = str(idempotency_key or "").strip()
        if not clean_subject_euid or not clean_idempotency_key:
            return None

        stmt = (
            select(action_instance)
            .where(
                action_instance.category == "action",
                action_instance.type == EXECUTION_ACTION_GROUP,
                action_instance.subtype == action_key,
                action_instance.is_deleted.is_(False),
                func.jsonb_extract_path_text(action_instance.json_addl, "subject_lookup_euid")
                == clean_subject_euid,
                func.jsonb_extract_path_text(action_instance.json_addl, "idempotency_key")
                == clean_idempotency_key,
            )
            .order_by(action_instance.created_dt.desc())
            .limit(1)
        )
        return self.session.execute(stmt).scalar_one_or_none()

    def record(
        self,
        *,
        target_instance,
        action_key: str,
        captured_data: dict[str, Any],
        result: dict[str, Any],
        executed_by: str | None,
        subject_euid: str | None,
        worker_euid: str | None,
        lease_euid: str | None,
        idempotency_key: str | None,
        payload_hash: str | None,
    ) -> action_instance:
        if action_key not in _ACTION_TEMPLATE_DEFINITIONS:
            raise ValueError(f"Unsupported execution action key: {action_key}")

        template = self._ensure_action_template(action_key)
        now = datetime.now(UTC).isoformat()
        action_definition = {
            "action_group": EXECUTION_ACTION_GROUP,
            "action_key": action_key,
            "name": _ACTION_TEMPLATE_DEFINITIONS[action_key]["name"],
            "description": _ACTION_TEMPLATE_DEFINITIONS[action_key]["description"],
        }
        action_record = action_instance(
            name=f"{action_key}@{target_instance.euid}",
            polymorphic_discriminator="action_instance",
            category="action",
            type=EXECUTION_ACTION_GROUP,
            subtype=action_key,
            version="1.0",
            template_uid=template.uid,
            json_addl={
                "action_group": EXECUTION_ACTION_GROUP,
                "action_key": action_key,
                "action_definition": action_definition,
                "captured_data": captured_data,
                "result": result,
                "executed_by": executed_by,
                "executed_at": now,
                "subject_lookup_euid": str(subject_euid or "").strip(),
                "worker_lookup_euid": str(worker_euid or "").strip(),
                "lease_lookup_euid": str(lease_euid or "").strip(),
                "idempotency_key": str(idempotency_key or "").strip(),
                "payload_hash": str(payload_hash or "").strip(),
                "target_euid": target_instance.euid,
            },
            bstatus="completed",
        )
        self.session.add(action_record)
        self.session.flush()

        lineage = action_instance_lineage(
            name=f"{action_record.euid}->{target_instance.euid}",
            parent_type="action:execution_queue",
            child_type=(
                f"{target_instance.category}:{target_instance.type}:"
                f"{target_instance.subtype}:{target_instance.version}"
            ),
            relationship_type="executed_on",
            parent_instance_uid=action_record.uid,
            child_instance_uid=target_instance.uid,
            category="action",
            type="action",
            subtype="executed_on",
            version="1.0",
            polymorphic_discriminator="action_instance_lineage",
            bstatus="active",
            json_addl={"properties": {"executed_at": now}},
        )
        self.session.add(lineage)
        self.session.flush()
        return action_record

    def _ensure_action_template(self, action_key: str) -> generic_template:
        ensure_instance_prefix_sequence(self.session, EXECUTION_ACTION_TEMPLATE_PREFIX)
        template_code = f"action/{EXECUTION_ACTION_GROUP}/{action_key}/1.0/"
        category, type_name, subtype, version = _parse_template_code(template_code)
        stmt = (
            select(generic_template)
            .where(
                generic_template.category == category,
                generic_template.type == type_name,
                generic_template.subtype == subtype,
                generic_template.version == version,
            )
            .limit(1)
        )
        existing = self.session.execute(stmt).scalar_one_or_none()
        if existing is not None:
            if existing.is_deleted:
                existing.is_deleted = False
            if existing.instance_prefix != EXECUTION_ACTION_TEMPLATE_PREFIX:
                existing.instance_prefix = EXECUTION_ACTION_TEMPLATE_PREFIX
            return existing

        template = generic_template(
            name=_ACTION_TEMPLATE_DEFINITIONS[action_key]["name"],
            polymorphic_discriminator="action_template",
            category=category,
            type=type_name,
            subtype=subtype,
            version=version,
            instance_prefix=EXECUTION_ACTION_TEMPLATE_PREFIX,
            instance_polymorphic_identity="action_instance",
            json_addl={
                "managed_by": "bloom",
                "domain": EXECUTION_ACTION_GROUP,
                "action_definition": {
                    "action_group": EXECUTION_ACTION_GROUP,
                    "action_key": action_key,
                    "name": _ACTION_TEMPLATE_DEFINITIONS[action_key]["name"],
                    "description": _ACTION_TEMPLATE_DEFINITIONS[action_key]["description"],
                },
            },
            bstatus="active",
            is_singleton=False,
            is_deleted=False,
        )
        self.session.add(template)
        self.session.flush()
        return template



"""Manual Atlas bridge APIs for query/status integration flows."""

from __future__ import annotations

import logging

from fastapi import APIRouter, Depends, Header, HTTPException

from bloom_lims.api.v1.dependencies import APIUser, require_external_atlas_api_enabled
from bloom_lims.auth.rbac import Permission
from bloom_lims.integrations.atlas.service import AtlasDependencyError, AtlasService
from bloom_lims.schemas.atlas_bridge import (
    AtlasTestStatusEventRequest,
    AtlasTestStatusEventResponse,
)


logger = logging.getLogger(__name__)

router = APIRouter(prefix="/external/atlas", tags=["External Atlas"])


def require_external_write(user: APIUser = Depends(require_external_atlas_api_enabled)) -> APIUser:
    if not user.has_permission(Permission.BLOOM_WRITE):
        raise HTTPException(status_code=403, detail="Write permission required")
    return user


@router.post(
    "/tests/{test_euid}/status-events",
    response_model=AtlasTestStatusEventResponse,
)
async def push_test_status_event(
    test_euid: str,
    payload: AtlasTestStatusEventRequest,
    user: APIUser = Depends(require_external_write),
    idempotency_key: str | None = Header(None, alias="Idempotency-Key"),
):
    _ = user
    service = AtlasService()
    try:
        result = service.push_test_status_event(
            test_euid=test_euid,
            payload=payload.model_dump(mode="json"),
            idempotency_key=idempotency_key,
        )
        return AtlasTestStatusEventResponse(**result.payload)
    except ValueError as exc:
        raise HTTPException(status_code=400, detail=str(exc)) from exc
    except AtlasDependencyError as exc:
        raise HTTPException(status_code=424, detail=str(exc)) from exc
    except Exception as exc:
        logger.exception("Failed pushing Atlas test status event")
        raise HTTPException(status_code=500, detail=str(exc)) from exc



"""
Pydantic schemas for workflows and workflow steps in BLOOM LIMS.
"""

from datetime import datetime
from typing import Any, Dict, List, Optional
from pydantic import Field, field_validator, model_validator

from .base import BloomBaseSchema, TimestampMixin, validate_euid


class WorkflowStepSchema(BloomBaseSchema):
    """Schema for workflow step definition."""
    
    name: str = Field(..., min_length=1, max_length=200, description="Step name")
    step_number: int = Field(..., ge=1, description="Step order number")
    description: Optional[str] = Field(None, max_length=2000, description="Step description")
    step_type: str = Field(default="standard", description="Type of step")
    required: bool = Field(default=True, description="Whether step is required")
    
    # Configuration
    config: Optional[Dict[str, Any]] = Field(default_factory=dict, description="Step configuration")
    input_schema: Optional[Dict[str, Any]] = Field(None, description="Expected input schema")
    output_schema: Optional[Dict[str, Any]] = Field(None, description="Expected output schema")
    
    # Conditions
    preconditions: Optional[List[str]] = Field(None, description="Preconditions to check")
    skip_conditions: Optional[List[str]] = Field(None, description="Conditions to skip step")


class WorkflowStepUpdateSchema(BloomBaseSchema):
    """Schema for updating a workflow step."""
    
    euid: str = Field(..., description="Step EUID to update")
    status: Optional[str] = Field(None, description="New status")
    completed_at: Optional[datetime] = Field(None, description="Completion timestamp")
    completed_by: Optional[str] = Field(None, description="User who completed")
    result: Optional[Dict[str, Any]] = Field(None, description="Step result data")
    notes: Optional[str] = Field(None, max_length=5000, description="Step notes")
    
    @field_validator("euid", mode="before")
    @classmethod
    def validate_euid(cls, v):
        """Validate step EUID."""
        return validate_euid(v)


class WorkflowBaseSchema(BloomBaseSchema):
    """Base schema for workflows."""
    
    name: str = Field(..., min_length=1, max_length=300, description="Workflow name")
    workflow_type: str = Field(..., min_length=1, max_length=100, description="Workflow type")
    description: Optional[str] = Field(None, max_length=5000, description="Workflow description")
    
    # Configuration
    config: Optional[Dict[str, Any]] = Field(default_factory=dict, description="Workflow config")
    
    @field_validator("workflow_type", mode="before")
    @classmethod
    def normalize_workflow_type(cls, v):
        """Normalize workflow type to lowercase."""
        if v:
            return str(v).strip().lower()
        return v


class WorkflowCreateSchema(WorkflowBaseSchema):
    """Schema for creating a new workflow."""
    
    steps: List[WorkflowStepSchema] = Field(
        default_factory=list,
        description="Workflow steps definition"
    )
    template_euid: Optional[str] = Field(None, description="Template to create from")
    
    # Objects to process
    object_euids: Optional[List[str]] = Field(None, description="Objects to attach to workflow")
    
    @field_validator("template_euid", mode="before")
    @classmethod
    def validate_template_euid(cls, v):
        """Validate template EUID if provided."""
        if v is not None and v.strip():
            return validate_euid(v)
        return None
    
    @field_validator("object_euids", mode="before")
    @classmethod
    def validate_object_euids(cls, v):
        """Validate object EUIDs if provided."""
        if v is None:
            return None
        validated = []
        for euid in v:
            if euid and euid.strip():
                validated.append(validate_euid(euid))
        return validated if validated else None
    
    @model_validator(mode="after")
    def validate_steps_or_template(self):
        """Ensure either steps or template is provided."""
        if not self.steps and not self.template_euid:
            raise ValueError("Either 'steps' or 'template_euid' must be provided")
        return self


class WorkflowStepResponseSchema(WorkflowStepSchema, TimestampMixin):
    """Schema for workflow step in API responses."""
    
    euid: str = Field(..., description="Step EUID")
    status: str = Field(default="pending", description="Step status")
    workflow_euid: str = Field(..., description="Parent workflow EUID")
    
    # Execution info
    started_at: Optional[datetime] = Field(None, description="Start timestamp")
    completed_at: Optional[datetime] = Field(None, description="Completion timestamp")
    completed_by: Optional[str] = Field(None, description="User who completed")
    result: Optional[Dict[str, Any]] = Field(None, description="Step result data")
    
    # Timing
    duration_seconds: Optional[int] = Field(None, description="Execution duration")


class WorkflowResponseSchema(WorkflowBaseSchema, TimestampMixin):
    """Schema for workflow API responses."""
    
    euid: str = Field(..., description="Workflow EUID")
    status: str = Field(default="pending", description="Workflow status")
    
    # Progress tracking
    current_step: Optional[int] = Field(None, description="Current step number")
    total_steps: int = Field(default=0, description="Total number of steps")
    completed_steps: int = Field(default=0, description="Number of completed steps")
    
    # Steps
    steps: List[WorkflowStepResponseSchema] = Field(default_factory=list, description="Workflow steps")
    
    # Timing
    started_at: Optional[datetime] = Field(None, description="Start timestamp")
    completed_at: Optional[datetime] = Field(None, description="Completion timestamp")
    
    @property
    def progress_percent(self) -> float:
        """Calculate completion percentage."""
        if self.total_steps == 0:
            return 0.0
        return (self.completed_steps / self.total_steps) * 100




"""
BLOOM LIMS API v1 - Containers Endpoints

CRUD endpoints for container management (plates, racks, boxes, etc.).
"""

import csv
import io
import json as json_lib
import logging
from typing import Any, Dict, List, Optional

from fastapi import APIRouter, Depends, File, HTTPException, Query, UploadFile

from bloom_lims.db import get_parent_lineages
from bloom_lims.integrations.atlas.events import emit_bloom_event
from bloom_lims.schemas import (
    ContainerCreateSchema,
    ContainerUpdateSchema,
    ContainerResponseSchema,
    PlaceInContainerSchema,
    BulkPlaceInContainerSchema,
    PaginatedResponse,
    SuccessResponse,
)
from bloom_lims.exceptions import NotFoundError, ValidationError
from .dependencies import APIUser, require_read, require_write


logger = logging.getLogger(__name__)

router = APIRouter(prefix="/containers", tags=["Containers"])


def get_bdb(username: str = "api-user"):
    """Get database connection."""
    from bloom_lims.db import BLOOMdb3
    return BLOOMdb3(app_username=username)


def _container_event_payload(container, extra: dict[str, Any] | None = None) -> dict[str, Any]:
    payload = {
        "euid": container.euid,
        "container_euid": container.euid,
        "name": container.name,
        "category": container.category,
        "type": container.type,
        "subtype": container.subtype,
        "status": container.bstatus,
        "is_deleted": bool(getattr(container, "is_deleted", False)),
    }
    json_addl = container.json_addl
    if isinstance(json_addl, str):
        try:
            json_addl = json_lib.loads(json_addl)
        except Exception:
            json_addl = {}
    payload["json_addl"] = json_addl if isinstance(json_addl, dict) else {}
    if extra:
        payload.update(extra)
    return payload


@router.get("/", response_model=Dict[str, Any])
async def list_containers(
    container_type: Optional[str] = Query(None, description="Filter by type (plate, rack, box)"),
    subtype: Optional[str] = Query(None, description="Filter by subtype"),
    status: Optional[str] = Query(None, description="Filter by status"),
    page: int = Query(1, ge=1),
    page_size: int = Query(50, ge=1, le=1000),
    user: APIUser = Depends(require_read),
):
    """List containers with optional filters."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomContainer

        bc = BloomContainer(bdb)
        bc.set_actor_context(user_id=user.user_id, email=user.email)
        query = bdb.session.query(bdb.Base.classes.generic_instance)
        query = query.filter(bdb.Base.classes.generic_instance.category == "container")

        if container_type:
            query = query.filter(bdb.Base.classes.generic_instance.type == container_type.lower())
        if subtype:
            query = query.filter(bdb.Base.classes.generic_instance.subtype == subtype.lower())
        if status:
            query = query.filter(bdb.Base.classes.generic_instance.bstatus == status)

        query = query.filter(bdb.Base.classes.generic_instance.is_deleted == False)

        total = query.count()
        offset = (page - 1) * page_size
        items = query.limit(page_size).offset(offset).all()

        return {
            "items": [
                {
                    "euid": obj.euid,
                    "name": obj.name,
                    "container_type": obj.type,
                    "subtype": obj.subtype,
                    "status": obj.bstatus,
                }
                for obj in items
            ],
            "total": total,
            "page": page,
            "page_size": page_size,
        }
    except Exception as e:
        logger.error(f"Error listing containers: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.get("/{euid}")
async def get_container(
    euid: str,
    include_contents: bool = Query(False),
    user: APIUser = Depends(require_read),
):
    """Get a container by EUID, optionally including contents."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomContainer
        
        bc = BloomContainer(bdb)
        container = bc.get_by_euid(euid)
        
        if not container:
            raise HTTPException(status_code=404, detail=f"Container not found: {euid}")
        
        result = {
            "euid": container.euid,
            "name": container.name,
            "container_type": container.type,
            "subtype": container.subtype,
            "status": container.bstatus,
            "json_addl": container.json_addl,
        }

        if include_contents:
            contents = []
            for lineage in get_parent_lineages(container):
                child = lineage.child_instance
                contents.append({
                    "euid": child.euid,
                    "name": child.name,
                    "type": child.type,
                    "position": child.json_addl.get("cont_address") if child.json_addl else None,
                })
            result["contents"] = contents
        
        return result
    except HTTPException:
        raise
    except Exception as e:
        error_msg = str(e)
        logger.error(f"Error getting container {euid}: {error_msg}")
        # Check for "not found" type errors and return 404
        if "not found" in error_msg.lower() or "no template found" in error_msg.lower():
            raise HTTPException(status_code=404, detail=f"Container not found: {euid}")
        raise HTTPException(status_code=500, detail=error_msg)


@router.post("/", response_model=Dict[str, Any])
async def create_container(
    data: ContainerCreateSchema,
    user: APIUser = Depends(require_write),
):
    """Create a new container from a template."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomContainer

        bc = BloomContainer(bdb)

        if data.template_euid:
            result = bc.create_empty_container(data.template_euid)
            container = result[0][0] if isinstance(result, list) else result
        else:
            raise HTTPException(status_code=400, detail="template_euid is required")

        emit_bloom_event("container.created", _container_event_payload(container))
        bc.track_user_interaction(
            container.euid,
            relationship_type="user_created",
            user_id=user.user_id,
            email=user.email,
        )

        return {
            "success": True,
            "euid": container.euid,
            "message": "Container created successfully",
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error creating container: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.put("/{euid}", response_model=Dict[str, Any])
async def update_container(
    euid: str,
    data: ContainerUpdateSchema,
    user: APIUser = Depends(require_write),
):
    """Update a container."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomContainer
        from sqlalchemy.orm.attributes import flag_modified

        bc = BloomContainer(bdb)
        bc.set_actor_context(user_id=user.user_id, email=user.email)
        container = bc.get_by_euid(euid)

        if not container:
            raise HTTPException(status_code=404, detail=f"Container not found: {euid}")

        prev_status = container.bstatus

        if data.name is not None:
            container.name = data.name
        if data.status is not None:
            container.bstatus = data.status
        if data.is_deleted is not None:
            container.is_deleted = data.is_deleted

        json_addl_dirty = False
        existing = container.json_addl or {}
        if isinstance(existing, str):
            try:
                existing = json_lib.loads(existing)
            except Exception:
                existing = {}
        if not isinstance(existing, dict):
            existing = {}

        if data.json_addl is not None:
            existing.update(data.json_addl)
            json_addl_dirty = True

        if data.metadata is not None:
            props = existing.get("properties")
            if not isinstance(props, dict):
                props = {}
            existing_metadata = props.get("metadata")
            if not isinstance(existing_metadata, dict):
                existing_metadata = {}
            merged_metadata = dict(existing_metadata)
            merged_metadata.update(data.metadata)
            props["metadata"] = merged_metadata
            existing["properties"] = props
            json_addl_dirty = True

        if json_addl_dirty:
            container.json_addl = existing
            flag_modified(container, "json_addl")

        bdb.session.commit()
        bc.track_user_interaction(
            container.euid,
            relationship_type="user_updated",
            user_id=user.user_id,
            email=user.email,
        )
        payload = _container_event_payload(container)
        emit_bloom_event("container.updated", payload)
        if prev_status != container.bstatus:
            emit_bloom_event(
                "container.status_changed",
                _container_event_payload(
                    container,
                    {
                        "previous_status": prev_status,
                        "current_status": container.bstatus,
                    },
                ),
            )

        return {
            "success": True,
            "euid": container.euid,
            "message": "Container updated successfully",
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error updating container {euid}: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.delete("/{euid}", response_model=Dict[str, Any])
async def delete_container(
    euid: str,
    hard_delete: bool = Query(False, description="Permanently delete"),
    user: APIUser = Depends(require_write),
):
    """Delete a container (soft delete by default)."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomContainer

        bc = BloomContainer(bdb)
        container = bc.get_by_euid(euid)

        if not container:
            raise HTTPException(status_code=404, detail=f"Container not found: {euid}")

        event_payload = _container_event_payload(
            container,
            {"hard_delete": hard_delete, "is_deleted": True},
        )
        if hard_delete:
            bc.delete_obj(container)
        else:
            container.is_deleted = True
            bdb.session.commit()

        emit_bloom_event("container.deleted", event_payload)

        return {
            "success": True,
            "euid": euid,
            "message": f"Container {'permanently deleted' if hard_delete else 'soft deleted'}",
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error deleting container {euid}: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.post("/{container_euid}/contents")
async def add_content_to_container(
    container_euid: str,
    data: PlaceInContainerSchema,
    user: APIUser = Depends(require_write),
):
    """Add content to a container."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomContainer

        bc = BloomContainer(bdb)
        # Schema field is `object_euid`; this endpoint is used for content placement.
        bc.link_content(container_euid, data.object_euid)

        return {"success": True, "message": "Content added to container"}
    except Exception as e:
        logger.error(f"Error adding content to container: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.delete("/{container_euid}/contents/{content_euid}", response_model=Dict[str, Any])
async def remove_content_from_container(
    container_euid: str,
    content_euid: str,
    user: APIUser = Depends(require_write),
):
    """Remove content from a container."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomContainer

        bc = BloomContainer(bdb)
        container = bc.get_by_euid(container_euid)

        if not container:
            raise HTTPException(status_code=404, detail=f"Container not found: {container_euid}")

        # Find and remove the lineage
        for lineage in get_parent_lineages(container):
            if lineage.child_instance.euid == content_euid and not lineage.is_deleted:
                lineage.is_deleted = True
                bdb.session.commit()
                return {"success": True, "message": "Content removed from container"}

        raise HTTPException(status_code=404, detail=f"Content {content_euid} not found in container")

    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error removing content from container: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.get("/{euid}/layout", response_model=Dict[str, Any])
async def get_container_layout(
    euid: str,
    user: APIUser = Depends(require_read),
):
    """Get the layout/wells of a container (for plates)."""
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomContainer

        bc = BloomContainer(bdb)
        container = bc.get_by_euid(euid)

        if not container:
            raise HTTPException(status_code=404, detail=f"Container not found: {euid}")

        layout = {}
        for lineage in get_parent_lineages(container):
            if lineage.is_deleted:
                continue
            child = lineage.child_instance
            if child.type == "well":
                addr = child.json_addl.get("cont_address", {}) if child.json_addl else {}
                position = addr.get("name", child.name)
                layout[position] = {
                    "euid": child.euid,
                    "name": child.name,
                    "status": child.bstatus,
                    "contents": [],
                }
                # Get well contents
                for well_lineage in get_parent_lineages(child):
                    if well_lineage.is_deleted:
                        continue
                    content = well_lineage.child_instance
                    if content.category == "content":
                        layout[position]["contents"].append({
                            "euid": content.euid,
                            "name": content.name,
                            "type": content.type,
                        })

        return {
            "container_euid": euid,
            "container_type": container.type,
            "layout": layout,
            "well_count": len(layout),
        }
    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error getting container layout {euid}: {e}")
        raise HTTPException(status_code=500, detail=str(e))


@router.post("/bulk-create", response_model=Dict[str, Any])
async def bulk_create_containers(
    file: UploadFile = File(..., description="TSV file with container/content data"),
    user: APIUser = Depends(require_write),
):
    """
    Bulk create containers with content from a TSV file.

    TSV columns:
    - container_template_euid OR container_type (e.g., "tube:tube-generic-10ml:1.0")
    - container_name (optional)
    - content_template_euid OR content_type (e.g., "sample:sample-blood:1.0")
    - content_name (optional)
    - content_properties (optional, JSON string for json_addl overrides)
    """
    try:
        bdb = get_bdb(user.email)
        from bloom_lims.bobjs import BloomObj

        bobj = BloomObj(bdb)

        # Read and parse TSV
        contents = await file.read()
        text = contents.decode("utf-8")
        reader = csv.DictReader(io.StringIO(text), delimiter="\t")

        results = {"created": [], "errors": [], "total_rows": 0}

        for row_num, row in enumerate(reader, start=2):
            results["total_rows"] += 1
            try:
                # Get container template
                cx_euid = row.get("container_template_euid", "").strip()
                cx_type = row.get("container_type", "").strip()

                if cx_euid:
                    cx_template = bobj.get_by_euid(cx_euid)
                    if not cx_template:
                        raise ValueError(f"Container template not found: {cx_euid}")
                elif cx_type:
                    # Parse type string: "type:subtype:version" or "type:subtype"
                    parts = cx_type.split(":")
                    if len(parts) < 2:
                        raise ValueError(f"Invalid container_type format: {cx_type}")
                    type_val, subtype_val = parts[0], parts[1]
                    version = parts[2] if len(parts) > 2 else "1.0"
                    templates = bobj.query_template_by_component_v2(
                        "container", type_val, subtype_val, version
                    )
                    if not templates:
                        raise ValueError(f"No container template found: {cx_type}")
                    cx_template = templates[0]
                else:
                    raise ValueError("container_template_euid or container_type required")

                # Get content template
                mx_euid = row.get("content_template_euid", "").strip()
                mx_type = row.get("content_type", "").strip()

                mx_template = None
                if mx_euid:
                    mx_template = bobj.get_by_euid(mx_euid)
                    if not mx_template:
                        raise ValueError(f"Content template not found: {mx_euid}")
                elif mx_type:
                    parts = mx_type.split(":")
                    if len(parts) < 2:
                        raise ValueError(f"Invalid content_type format: {mx_type}")
                    type_val, subtype_val = parts[0], parts[1]
                    version = parts[2] if len(parts) > 2 else "1.0"
                    templates = bobj.query_template_by_component_v2(
                        "content", type_val, subtype_val, version
                    )
                    if not templates:
                        raise ValueError(f"No content template found: {mx_type}")
                    mx_template = templates[0]

                # Create container
                container_result = bobj.create_instances(cx_template.euid)
                container = container_result[0][0] if isinstance(container_result, list) else container_result

                # Set container name if provided
                cx_name = row.get("container_name", "").strip()
                if cx_name:
                    container.name = cx_name
                    from sqlalchemy.orm.attributes import flag_modified
                    if container.json_addl and "properties" in container.json_addl:
                        container.json_addl["properties"]["name"] = cx_name
                        flag_modified(container, "json_addl")

                content = None
                if mx_template:
                    # Parse content properties
                    props_str = row.get("content_properties", "").strip()
                    json_overrides = {}
                    if props_str:
                        try:
                            json_overrides = json_lib.loads(props_str)
                        except json_lib.JSONDecodeError:
                            raise ValueError(f"Invalid JSON in content_properties: {props_str}")

                    # Create content
                    content = bobj.create_instance(mx_template.euid, json_overrides)

                    # Set content name if provided
                    mx_name = row.get("content_name", "").strip()
                    if mx_name:
                        content.name = mx_name
                        if content.json_addl and "properties" in content.json_addl:
                            content.json_addl["properties"]["name"] = mx_name
                            flag_modified(content, "json_addl")

                    # Link content to container
                    from bloom_lims.bobjs import BloomContainer
                    bc = BloomContainer(bdb)
                    bc.link_content(container.euid, content.euid)

                bdb.session.commit()

                results["created"].append({
                    "row": row_num,
                    "container_euid": container.euid,
                    "content_euid": content.euid if content else None,
                })

            except Exception as e:
                results["errors"].append({
                    "row": row_num,
                    "error": str(e),
                })

        return {
            "success": True,
            "total_rows": results["total_rows"],
            "created_count": len(results["created"]),
            "error_count": len(results["errors"]),
            "created": results["created"],
            "errors": results["errors"],
        }

    except HTTPException:
        raise
    except Exception as e:
        logger.error(f"Error in bulk container creation: {e}")
        raise HTTPException(status_code=500, detail=str(e))



"""
BLOOM LIMS Base Objects Module

This module contains the BloomObj base class and common utility functions
for working with BLOOM LIMS objects.

For backward compatibility, this module re-exports functionality that was
originally in bloom_lims/bobjs.py.
"""

import logging
from typing import Any, Dict, List, Optional, Type, Union
from datetime import datetime

from sqlalchemy import and_, desc, or_
from sqlalchemy.orm import Session

from bloom_lims.exceptions import (
    NotFoundError,
    ValidationError,
    DatabaseError,
)


logger = logging.getLogger(__name__)


class BloomObjMixin:
    """
    Mixin class providing common functionality for BLOOM objects.
    
    This mixin can be used to add standard BLOOM object methods to
    any SQLAlchemy model class.
    """
    
    @property
    def is_template(self) -> bool:
        """Check if this object is a template."""
        return hasattr(self, 'polymorphic_discriminator') and \
               self.polymorphic_discriminator and \
               '_template' in self.polymorphic_discriminator
    
    @property
    def is_instance(self) -> bool:
        """Check if this object is an instance."""
        return hasattr(self, 'polymorphic_discriminator') and \
               self.polymorphic_discriminator and \
               '_instance' in self.polymorphic_discriminator
    
    def get_json_addl_value(self, key: str, default: Any = None) -> Any:
        """
        Safely get a value from json_addl.
        
        Args:
            key: The key to look up
            default: Default value if key not found
            
        Returns:
            The value or default
        """
        if hasattr(self, 'json_addl') and self.json_addl:
            return self.json_addl.get(key, default)
        return default
    
    def set_json_addl_value(self, key: str, value: Any) -> None:
        """
        Set a value in json_addl.
        
        Args:
            key: The key to set
            value: The value to set
        """
        if not hasattr(self, 'json_addl') or self.json_addl is None:
            self.json_addl = {}
        self.json_addl[key] = value
    
    def merge_json_addl(self, data: Dict[str, Any]) -> None:
        """
        Merge data into json_addl recursively.
        
        Args:
            data: Dictionary to merge into json_addl
        """
        if not hasattr(self, 'json_addl') or self.json_addl is None:
            self.json_addl = {}
        _update_recursive(self.json_addl, data)


def _update_recursive(orig_dict: Dict, update_with: Dict) -> None:
    """
    Recursively update a dictionary.
    
    Args:
        orig_dict: Original dictionary to update
        update_with: Dictionary with updates
    """
    for key, value in update_with.items():
        if (
            key in orig_dict
            and isinstance(orig_dict[key], dict)
            and isinstance(value, dict)
        ):
            _update_recursive(orig_dict[key], value)
        else:
            orig_dict[key] = value


def create_bloom_obj(
    session: Session,
    base,
    obj_class: str,
    name: str,
    type: str,
    subtype: Optional[str] = None,
    json_addl: Optional[Dict[str, Any]] = None,
    template_uid: Optional[str] = None,
    **kwargs,
) -> Any:
    """
    Create a new BloomObj instance.

    Args:
        session: SQLAlchemy session
        base: SQLAlchemy automap base
        obj_class: Object class name (e.g., 'content_instance')
        name: Object name
        type: Object type
        subtype: Object subtype (optional)
        json_addl: Additional JSON data (optional)
        template_uid: Template UID if creating from template (optional)
        **kwargs: Additional fields to set

    Returns:
        The created object

    Raises:
        ValidationError: If required fields are missing
        DatabaseError: If database operation fails
    """
    logger.debug(f"Creating {obj_class} with name={name}, type={type}")

    if not name or not type:
        raise ValidationError("name and type are required", field="name,type")

    try:
        obj_class_ref = getattr(base.classes, obj_class)
        obj = obj_class_ref(
            name=name,
            type=type,
            subtype=subtype,
            json_addl=json_addl or {},
            template_uid=template_uid,
            **kwargs,
        )
        session.add(obj)
        session.flush()
        return obj
    except AttributeError as e:
        raise ValidationError(f"Invalid object class: {obj_class}", field="obj_class")
    except Exception as e:
        logger.error(f"Error creating {obj_class}: {e}")
        raise DatabaseError(f"Failed to create {obj_class}: {e}", operation="insert")


def get_bloom_obj_by_euid(
    session: Session,
    base,
    euid: str,
    include_deleted: bool = False,
) -> Optional[Any]:
    """
    Get a BloomObj by its EUID.

    Searches across all generic_instance types for the given EUID.

    Args:
        session: SQLAlchemy session
        base: SQLAlchemy automap base
        euid: Entity Unique Identifier
        include_deleted: Include soft-deleted objects

    Returns:
        The object if found, None otherwise
    """
    logger.debug(f"Looking up object by EUID: {euid}")

    if not euid:
        return None

    euid = euid.strip().upper()

    try:
        query = session.query(base.classes.generic_instance).filter(
            base.classes.generic_instance.euid == euid
        )

        if not include_deleted:
            query = query.filter(
                base.classes.generic_instance.is_deleted == False
            )

        return query.first()
    except Exception as e:
        logger.error(f"Error looking up EUID {euid}: {e}")
        return None


def query_objects(
    session: Session,
    base,
    type: Optional[str] = None,
    subtype: Optional[str] = None,
    category: Optional[str] = None,
    status: Optional[str] = None,
    include_deleted: bool = False,
    limit: int = 100,
    offset: int = 0,
) -> List[Any]:
    """
    Query BloomObj instances with filters.

    Args:
        session: SQLAlchemy session
        base: SQLAlchemy automap base
        type: Filter by type
        subtype: Filter by subtype
        category: Filter by category
        status: Filter by status
        include_deleted: Include soft-deleted objects
        limit: Maximum results
        offset: Results offset

    Returns:
        List of matching objects
    """
    logger.debug(f"Querying objects: type={type}, subtype={subtype}")

    try:
        query = session.query(base.classes.generic_instance)

        if type:
            query = query.filter(base.classes.generic_instance.type == type.lower())

        if subtype:
            query = query.filter(base.classes.generic_instance.subtype == subtype.lower())

        if category:
            query = query.filter(base.classes.generic_instance.category == category)

        if status:
            query = query.filter(base.classes.generic_instance.bstatus == status)

        if not include_deleted:
            query = query.filter(base.classes.generic_instance.is_deleted == False)

        return query.limit(limit).offset(offset).all()
    except Exception as e:
        logger.error(f"Error querying objects: {e}")
        return []


# Re-export BloomObj for backward compatibility
# The actual BloomObj class is still in bobjs.py for now
# This reference will be updated when full migration is complete
try:
    from bloom_lims.bobjs import BloomObj
except ImportError:
    # During initial import, bobjs may not be available yet
    BloomObj = None




from __future__ import annotations

from datetime import UTC, datetime

from bloom_lims.gui.jinja import templates


def _render(template_body: str, **kwargs) -> str:
    template = templates.from_string(template_body)
    return template.render(**kwargs)


def test_format_dt_converts_utc_to_user_display_timezone():
    rendered = _render(
        "{{ value | format_dt('standard') }}",
        value=datetime(2026, 3, 9, 12, 0, 0, tzinfo=UTC),
        udat={"display_timezone": "America/New_York"},
    )
    assert rendered.startswith("2026-03-09 08:00:00")


def test_format_dt_defaults_to_utc_for_invalid_timezone():
    rendered = _render(
        "{{ value | format_dt('standard') }}",
        value=datetime(2026, 3, 9, 12, 0, 0, tzinfo=UTC),
        udat={"display_timezone": "Not/A_Timezone"},
    )
    assert rendered.startswith("2026-03-09 12:00:00")


def test_dt_local_input_treats_naive_timestamp_as_utc():
    rendered = _render(
        "{{ value | dt_local_input }}",
        value=datetime(2026, 1, 1, 0, 0, 0),
        udat={"display_timezone": "America/Los_Angeles"},
    )
    assert rendered == "2025-12-31T16:00:00"



#!/usr/bin/env python3
"""Provision or update a local Bloom user and group membership."""

from __future__ import annotations

import argparse
import json

from bloom_lims.auth.services.groups import GroupService
from bloom_lims.db import BLOOMdb3
from daylily_tapdb.user_store import create_or_get, set_role


def main() -> int:
    parser = argparse.ArgumentParser(description="Provision a local Bloom user")
    parser.add_argument("--username", required=True)
    parser.add_argument("--email", default="")
    parser.add_argument("--name", required=True)
    parser.add_argument("--role", default="admin")
    parser.add_argument("--group", action="append", default=[])
    parser.add_argument("--json", action="store_true")
    args = parser.parse_args()

    username = str(args.username).strip().lower()
    email = str(args.email or args.username).strip().lower()
    role = str(args.role or "admin").strip().lower()
    groups = [str(group).strip().upper() for group in args.group if str(group).strip()]

    bdb = BLOOMdb3(app_username="local-provision")
    try:
        user, _created = create_or_get(
            bdb.session,
            login_identifier=username,
            email=email,
            display_name=args.name.strip(),
            role=role,
            cognito_username=email,
        )
        set_role(bdb.session, username, role)
        group_service = GroupService(bdb.session)
        group_service.ensure_system_groups()
        for group_code in groups:
            group_service.add_user_to_group(
                group_code=group_code,
                user_id=str(user.uid),
                added_by=str(user.uid),
            )
        bdb.session.commit()
        payload = {
            "uid": user.uid,
            "euid": user.euid,
            "username": user.username,
            "email": user.email,
            "role": role,
            "groups": groups,
        }
    except Exception as exc:
        bdb.session.rollback()
        print(f"Provision failed: {exc}")
        return 1
    finally:
        bdb.close()

    if args.json:
        print(json.dumps(payload, indent=2, sort_keys=True))
    else:
        print(json.dumps(payload, sort_keys=True))
    return 0


if __name__ == "__main__":
    raise SystemExit(main())



"""
BLOOM LIMS Custom Exception Hierarchy

This module provides a standardized exception hierarchy for the BLOOM LIMS system.
All custom exceptions inherit from BloomError, enabling consistent error handling
and logging throughout the application.

Usage:
    from bloom_lims.exceptions import ValidationError, NotFoundError
    
    if not is_valid_euid(euid):
        raise ValidationError(f"Invalid EUID format: {euid}", field="euid")
    
    obj = get_by_euid(euid)
    if obj is None:
        raise NotFoundError(f"Object not found", resource_type="instance", resource_id=euid)
"""

import logging
from typing import Any, Dict, Optional
from datetime import UTC, datetime


logger = logging.getLogger(__name__)


class BloomError(Exception):
    """
    Base exception for all BLOOM LIMS errors.
    
    Provides structured error information including:
    - Error code for programmatic handling
    - Detailed message for logging
    - Optional context data for debugging
    - Timestamp for error tracking
    
    Args:
        message: Human-readable error description
        error_code: Machine-readable error code (default: BLOOM_ERROR)
        details: Additional context data for debugging
        cause: Original exception that caused this error
    """
    
    default_error_code = "BLOOM_ERROR"
    default_http_status = 500
    
    def __init__(
        self,
        message: str,
        error_code: Optional[str] = None,
        details: Optional[Dict[str, Any]] = None,
        cause: Optional[Exception] = None,
    ):
        self.message = message
        self.error_code = error_code or self.default_error_code
        self.details = details or {}
        self.cause = cause
        self.timestamp = datetime.now(UTC).isoformat()
        self.http_status = self.default_http_status
        
        super().__init__(self.message)
        
        # Log the error with structured data
        self._log_error()
    
    def _log_error(self) -> None:
        """Log error with structured format."""
        log_data = {
            "error_code": self.error_code,
            "message": self.message,
            "timestamp": self.timestamp,
            "details": self.details,
        }
        if self.cause:
            log_data["cause"] = str(self.cause)
        
        logger.error(f"BloomError: {self.error_code} - {self.message}", extra=log_data)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert error to dictionary for API responses."""
        result = {
            "error": True,
            "error_code": self.error_code,
            "message": self.message,
            "timestamp": self.timestamp,
        }
        if self.details:
            result["details"] = self.details
        return result
    
    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(code={self.error_code}, message={self.message})"


class ValidationError(BloomError):
    """
    Raised when input validation fails.
    
    Args:
        message: Description of validation failure
        field: Name of the field that failed validation
        value: The invalid value (will be sanitized in logs)
        constraints: Description of validation constraints
    """
    
    default_error_code = "VALIDATION_ERROR"
    default_http_status = 400
    
    def __init__(
        self,
        message: str,
        field: Optional[str] = None,
        value: Optional[Any] = None,
        constraints: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if field:
            details["field"] = field
        if value is not None:
            # Sanitize value for logging (truncate long values)
            str_value = str(value)
            details["value"] = str_value[:100] + "..." if len(str_value) > 100 else str_value
        if constraints:
            details["constraints"] = constraints
        
        super().__init__(message, details=details, **kwargs)


class DatabaseError(BloomError):
    """
    Raised when database operations fail.

    Args:
        message: Description of database error
        operation: The database operation that failed (query, insert, update, delete)
        table: The table involved in the operation
    """

    default_error_code = "DATABASE_ERROR"
    default_http_status = 500

    def __init__(
        self,
        message: str,
        operation: Optional[str] = None,
        table: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if operation:
            details["operation"] = operation
        if table:
            details["table"] = table

        super().__init__(message, details=details, **kwargs)


class SessionError(DatabaseError):
    """
    Raised when database session management fails.

    Used for connection issues, session lifecycle problems, and transaction errors.
    """

    default_error_code = "SESSION_ERROR"
    default_http_status = 500


class AuthenticationError(BloomError):
    """
    Raised when authentication fails.

    Args:
        message: Description of authentication failure
        username: The username that failed authentication (optional)
        auth_method: The authentication method used
    """

    default_error_code = "AUTHENTICATION_ERROR"
    default_http_status = 401

    def __init__(
        self,
        message: str,
        username: Optional[str] = None,
        auth_method: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if username:
            details["username"] = username
        if auth_method:
            details["auth_method"] = auth_method

        super().__init__(message, details=details, **kwargs)


class AuthorizationError(BloomError):
    """
    Raised when user lacks required permissions.

    Args:
        message: Description of authorization failure
        user_id: The user who was denied access
        resource: The resource access was denied to
        required_permission: The permission that was required
    """

    default_error_code = "AUTHORIZATION_ERROR"
    default_http_status = 403

    def __init__(
        self,
        message: str,
        user_id: Optional[str] = None,
        resource: Optional[str] = None,
        required_permission: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if user_id:
            details["user_id"] = user_id
        if resource:
            details["resource"] = resource
        if required_permission:
            details["required_permission"] = required_permission

        super().__init__(message, details=details, **kwargs)


class NotFoundError(BloomError):
    """
    Raised when a requested resource cannot be found.

    Args:
        message: Description of what was not found
        resource_type: Type of resource (instance, template, lineage, etc.)
        resource_id: Identifier of the resource (EUID, UUID, etc.)
    """

    default_error_code = "NOT_FOUND"
    default_http_status = 404

    def __init__(
        self,
        message: str,
        resource_type: Optional[str] = None,
        resource_id: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if resource_type:
            details["resource_type"] = resource_type
        if resource_id:
            details["resource_id"] = resource_id

        super().__init__(message, details=details, **kwargs)


class DuplicateError(BloomError):
    """
    Raised when attempting to create a duplicate resource.

    Args:
        message: Description of duplicate conflict
        resource_type: Type of resource
        identifier: The conflicting identifier
    """

    default_error_code = "DUPLICATE_ERROR"
    default_http_status = 409

    def __init__(
        self,
        message: str,
        resource_type: Optional[str] = None,
        identifier: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if resource_type:
            details["resource_type"] = resource_type
        if identifier:
            details["identifier"] = identifier

        super().__init__(message, details=details, **kwargs)


class ConfigurationError(BloomError):
    """
    Raised when configuration is invalid or missing.

    Args:
        message: Description of configuration error
        config_key: The configuration key that is problematic
        expected: What was expected
    """

    default_error_code = "CONFIGURATION_ERROR"
    default_http_status = 500

    def __init__(
        self,
        message: str,
        config_key: Optional[str] = None,
        expected: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if config_key:
            details["config_key"] = config_key
        if expected:
            details["expected"] = expected

        super().__init__(message, details=details, **kwargs)


class WorkflowError(BloomError):
    """
    Raised when workflow operations fail.

    Args:
        message: Description of workflow error
        workflow_euid: EUID of the workflow
        step_euid: EUID of the workflow step (if applicable)
        action: The action that failed
    """

    default_error_code = "WORKFLOW_ERROR"
    default_http_status = 400

    def __init__(
        self,
        message: str,
        workflow_euid: Optional[str] = None,
        step_euid: Optional[str] = None,
        action: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if workflow_euid:
            details["workflow_euid"] = workflow_euid
        if step_euid:
            details["step_euid"] = step_euid
        if action:
            details["action"] = action

        super().__init__(message, details=details, **kwargs)


class StorageError(BloomError):
    """
    Raised when file/object storage operations fail.

    Args:
        message: Description of storage error
        storage_type: Type of storage (S3, local, etc.)
        path: The path or key that had issues
        operation: The operation that failed
    """

    default_error_code = "STORAGE_ERROR"
    default_http_status = 500

    def __init__(
        self,
        message: str,
        storage_type: Optional[str] = None,
        path: Optional[str] = None,
        operation: Optional[str] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if storage_type:
            details["storage_type"] = storage_type
        if path:
            details["path"] = path
        if operation:
            details["operation"] = operation

        super().__init__(message, details=details, **kwargs)


class ExternalServiceError(BloomError):
    """
    Raised when external service calls fail.

    Args:
        message: Description of external service error
        service_name: Name of the external service
        endpoint: The endpoint that was called
        status_code: HTTP status code from the service (if applicable)
    """

    default_error_code = "EXTERNAL_SERVICE_ERROR"
    default_http_status = 502

    def __init__(
        self,
        message: str,
        service_name: Optional[str] = None,
        endpoint: Optional[str] = None,
        status_code: Optional[int] = None,
        **kwargs,
    ):
        details = kwargs.pop("details", {})
        if service_name:
            details["service_name"] = service_name
        if endpoint:
            details["endpoint"] = endpoint
        if status_code is not None:
            details["status_code"] = status_code

        super().__init__(message, details=details, **kwargs)


# Convenience function for creating error responses in FastAPI
def create_error_response(error: BloomError) -> Dict[str, Any]:
    """
    Create a standardized error response dictionary for FastAPI.

    Args:
        error: The BloomError instance

    Returns:
        Dictionary suitable for JSONResponse
    """
    return {
        "success": False,
        "error": error.to_dict(),
    }



"""
BLOOM LIMS Batch Operations

Provides bulk processing capabilities for high-throughput operations.

Usage:
    from bloom_lims.core.batch_operations import (
        BatchProcessor,
        BatchJob,
        bulk_create,
        bulk_update,
        bulk_delete,
    )
    
    # Batch processing with progress tracking
    processor = BatchProcessor(bdb)
    job = await processor.create_objects(
        template_euid="CT123",
        count=1000,
        data_generator=lambda i: {"name": f"Sample_{i}"},
    )
    
    # Check progress
    status = processor.get_job_status(job.job_id)
    
    # Bulk operations
    results = await bulk_create(session, objects_data)
    updated = await bulk_update(session, updates)

Features:
    - Batched database operations for efficiency
    - Progress tracking for long-running jobs
    - Automatic chunking to prevent memory issues
    - Transaction management with rollback support
    - Parallel processing options
"""

import asyncio
import logging
import secrets
from dataclasses import dataclass, field
from datetime import UTC, datetime
from enum import Enum
from typing import Any, Callable, Dict, Generator, List, Optional, TypeVar, Union

from sqlalchemy import text, update, delete
from sqlalchemy.orm import Session

from bloom_lims.config import get_settings

logger = logging.getLogger(__name__)

T = TypeVar("T")


class JobStatus(str, Enum):
    """Status of a batch job."""
    PENDING = "pending"
    RUNNING = "running"
    COMPLETED = "completed"
    FAILED = "failed"
    CANCELLED = "cancelled"
    PAUSED = "paused"


@dataclass
class BatchProgress:
    """Progress tracking for batch operations."""
    total: int = 0
    processed: int = 0
    succeeded: int = 0
    failed: int = 0
    skipped: int = 0
    
    @property
    def percent_complete(self) -> float:
        if self.total == 0:
            return 0.0
        return (self.processed / self.total) * 100
    
    @property
    def remaining(self) -> int:
        return self.total - self.processed


@dataclass
class BatchJob:
    """Represents a batch processing job."""
    job_id: str
    operation: str
    status: JobStatus = JobStatus.PENDING
    progress: BatchProgress = field(default_factory=BatchProgress)
    created_at: datetime = field(default_factory=lambda: datetime.now(UTC))
    started_at: Optional[datetime] = None
    completed_at: Optional[datetime] = None
    error: Optional[str] = None
    results: List[Dict[str, Any]] = field(default_factory=list)
    errors: List[Dict[str, Any]] = field(default_factory=list)
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary."""
        return {
            "job_id": self.job_id,
            "operation": self.operation,
            "status": self.status.value,
            "progress": {
                "total": self.progress.total,
                "processed": self.progress.processed,
                "succeeded": self.progress.succeeded,
                "failed": self.progress.failed,
                "skipped": self.progress.skipped,
                "percent_complete": round(self.progress.percent_complete, 2),
            },
            "created_at": self.created_at.isoformat(),
            "started_at": self.started_at.isoformat() if self.started_at else None,
            "completed_at": self.completed_at.isoformat() if self.completed_at else None,
            "error": self.error,
            "result_count": len(self.results),
            "error_count": len(self.errors),
        }


class BatchProcessor:
    """
    Processor for batch database operations.
    
    Handles bulk creates, updates, and deletes with progress tracking.
    """
    
    def __init__(
        self,
        bdb,
        chunk_size: int = 100,
        max_concurrent: int = 5,
    ):
        """
        Initialize batch processor.
        
        Args:
            bdb: BLOOMdb3 database instance
            chunk_size: Number of items per batch
            max_concurrent: Maximum concurrent operations
        """
        self._bdb = bdb
        self._session = bdb.session
        self._Base = bdb.Base
        self.chunk_size = chunk_size
        self.max_concurrent = max_concurrent
        self._jobs: Dict[str, BatchJob] = {}
        self._cancel_flags: Dict[str, bool] = {}
    
    def _chunks(self, items: List[T], size: int) -> Generator[List[T], None, None]:
        """Split list into chunks."""
        for i in range(0, len(items), size):
            yield items[i:i + size]

    def create_job(self, operation: str, total: int) -> BatchJob:
        """Create a new batch job."""
        job = BatchJob(
            job_id=f"job_{secrets.token_hex(16)}",
            operation=operation,
            progress=BatchProgress(total=total),
        )
        self._jobs[job.job_id] = job
        self._cancel_flags[job.job_id] = False
        return job

    def get_job(self, job_id: str) -> Optional[BatchJob]:
        """Get job by ID."""
        return self._jobs.get(job_id)

    def cancel_job(self, job_id: str) -> bool:
        """Request job cancellation."""
        if job_id in self._cancel_flags:
            self._cancel_flags[job_id] = True
            return True
        return False

    def _is_cancelled(self, job_id: str) -> bool:
        """Check if job is cancelled."""
        return self._cancel_flags.get(job_id, False)

    async def bulk_create_objects(
        self,
        template_euid: str,
        count: int,
        data_generator: Optional[Callable[[int], Dict[str, Any]]] = None,
        name_pattern: str = "Object_{index}",
    ) -> BatchJob:
        """
        Create multiple objects from a template.

        Args:
            template_euid: Template EUID to instantiate
            count: Number of objects to create
            data_generator: Optional function to generate json_addl for each object
            name_pattern: Name pattern with {index} placeholder

        Returns:
            BatchJob with created object EUIDs
        """
        job = self.create_job("bulk_create", count)
        job.status = JobStatus.RUNNING
        job.started_at = datetime.now(UTC)

        try:
            # Get template
            template = (
                self._session.query(self._Base.classes.generic_template)
                .filter(self._Base.classes.generic_template.euid == template_euid)
                .first()
            )

            if not template:
                raise ValueError(f"Template not found: {template_euid}")

            # Process in chunks
            for chunk_start in range(0, count, self.chunk_size):
                if self._is_cancelled(job.job_id):
                    job.status = JobStatus.CANCELLED
                    break

                chunk_end = min(chunk_start + self.chunk_size, count)

                for i in range(chunk_start, chunk_end):
                    try:
                        # Generate object data
                        name = name_pattern.format(index=i + 1)
                        json_addl = data_generator(i) if data_generator else {}

                        # Create instance
                        instance_class = getattr(self._Base.classes, 'generic_instance')
                        instance = instance_class(
                            name=name,
                            type=template.type,
                            subtype=template.subtype,
                            category=template.category,
                            json_addl=json_addl,
                            template_uid=template.uid,
                        )
                        self._session.add(instance)

                        job.progress.succeeded += 1
                        job.results.append({"index": i, "name": name})

                    except Exception as e:
                        job.progress.failed += 1
                        job.errors.append({"index": i, "error": str(e)})
                        logger.warning(f"Failed to create object {i}: {e}")

                    job.progress.processed += 1

                # Commit chunk
                self._session.commit()
                logger.debug(f"Committed chunk {chunk_start}-{chunk_end}")

            job.status = JobStatus.COMPLETED

        except Exception as e:
            job.status = JobStatus.FAILED
            job.error = str(e)
            self._session.rollback()
            logger.error(f"Batch create failed: {e}")

        job.completed_at = datetime.now(UTC)
        return job

    async def bulk_update_objects(
        self,
        updates: List[Dict[str, Any]],
    ) -> BatchJob:
        """
        Update multiple objects by EUID.

        Args:
            updates: List of dicts with 'euid' and fields to update

        Returns:
            BatchJob with update results
        """
        job = self.create_job("bulk_update", len(updates))
        job.status = JobStatus.RUNNING
        job.started_at = datetime.now(UTC)

        try:
            for chunk in self._chunks(updates, self.chunk_size):
                if self._is_cancelled(job.job_id):
                    job.status = JobStatus.CANCELLED
                    break

                for item in chunk:
                    euid = item.get('euid')
                    if not euid:
                        job.progress.skipped += 1
                        job.progress.processed += 1
                        continue

                    try:
                        obj = (
                            self._session.query(self._Base.classes.generic_instance)
                            .filter(self._Base.classes.generic_instance.euid == euid)
                            .first()
                        )

                        if obj:
                            for key, value in item.items():
                                if key != 'euid' and hasattr(obj, key):
                                    setattr(obj, key, value)

                            job.progress.succeeded += 1
                            job.results.append({"euid": euid, "status": "updated"})
                        else:
                            job.progress.skipped += 1
                            job.errors.append({"euid": euid, "error": "not_found"})

                    except Exception as e:
                        job.progress.failed += 1
                        job.errors.append({"euid": euid, "error": str(e)})

                    job.progress.processed += 1

                self._session.commit()

            job.status = JobStatus.COMPLETED

        except Exception as e:
            job.status = JobStatus.FAILED
            job.error = str(e)
            self._session.rollback()

        job.completed_at = datetime.now(UTC)
        return job

    async def bulk_delete_objects(
        self,
        euids: List[str],
        soft_delete: bool = True,
    ) -> BatchJob:
        """
        Delete multiple objects by EUID.

        Args:
            euids: List of EUIDs to delete
            soft_delete: If True, set is_deleted=True; if False, hard delete

        Returns:
            BatchJob with deletion results
        """
        job = self.create_job("bulk_delete", len(euids))
        job.status = JobStatus.RUNNING
        job.started_at = datetime.now(UTC)

        try:
            for chunk in self._chunks(euids, self.chunk_size):
                if self._is_cancelled(job.job_id):
                    job.status = JobStatus.CANCELLED
                    break

                for euid in chunk:
                    try:
                        obj = (
                            self._session.query(self._Base.classes.generic_instance)
                            .filter(self._Base.classes.generic_instance.euid == euid)
                            .first()
                        )

                        if obj:
                            if soft_delete:
                                obj.is_deleted = True
                            else:
                                self._session.delete(obj)

                            job.progress.succeeded += 1
                            job.results.append({"euid": euid, "status": "deleted"})
                        else:
                            job.progress.skipped += 1

                    except Exception as e:
                        job.progress.failed += 1
                        job.errors.append({"euid": euid, "error": str(e)})

                    job.progress.processed += 1

                self._session.commit()

            job.status = JobStatus.COMPLETED

        except Exception as e:
            job.status = JobStatus.FAILED
            job.error = str(e)
            self._session.rollback()

        job.completed_at = datetime.now(UTC)
        return job

    def get_all_jobs(self) -> List[BatchJob]:
        """Get all batch jobs."""
        return list(self._jobs.values())

    def cleanup_completed_jobs(self, max_age_hours: int = 24) -> int:
        """Remove old completed jobs."""
        cutoff = datetime.now(UTC)
        removed = 0

        to_remove = []
        for job_id, job in self._jobs.items():
            if job.status in (JobStatus.COMPLETED, JobStatus.FAILED, JobStatus.CANCELLED):
                if job.completed_at:
                    age_hours = (cutoff - job.completed_at).total_seconds() / 3600
                    if age_hours > max_age_hours:
                        to_remove.append(job_id)

        for job_id in to_remove:
            del self._jobs[job_id]
            if job_id in self._cancel_flags:
                del self._cancel_flags[job_id]
            removed += 1

        return removed


# Convenience functions for simple bulk operations
async def bulk_create(
    session: Session,
    model_class,
    items: List[Dict[str, Any]],
    chunk_size: int = 100,
) -> Dict[str, Any]:
    """
    Bulk create objects directly.

    Args:
        session: SQLAlchemy session
        model_class: ORM model class
        items: List of dictionaries with object data
        chunk_size: Items per transaction chunk

    Returns:
        Summary dict with counts and any errors
    """
    results = {"created": 0, "failed": 0, "errors": []}

    for i in range(0, len(items), chunk_size):
        chunk = items[i:i + chunk_size]

        try:
            for item in chunk:
                obj = model_class(**item)
                session.add(obj)
                results["created"] += 1

            session.commit()
        except Exception as e:
            session.rollback()
            results["failed"] += len(chunk)
            results["errors"].append({"chunk": i // chunk_size, "error": str(e)})

    return results


async def bulk_update(
    session: Session,
    model_class,
    updates: List[Dict[str, Any]],
    id_field: str = "euid",
    chunk_size: int = 100,
) -> Dict[str, Any]:
    """
    Bulk update objects directly.

    Args:
        session: SQLAlchemy session
        model_class: ORM model class
        updates: List of dicts with id_field and fields to update
        id_field: Field name used for lookup
        chunk_size: Items per transaction chunk

    Returns:
        Summary dict with counts
    """
    results = {"updated": 0, "not_found": 0, "failed": 0, "errors": []}

    for i in range(0, len(updates), chunk_size):
        chunk = updates[i:i + chunk_size]

        try:
            for item in chunk:
                id_value = item.get(id_field)
                if not id_value:
                    continue

                obj = session.query(model_class).filter(
                    getattr(model_class, id_field) == id_value
                ).first()

                if obj:
                    for key, value in item.items():
                        if key != id_field and hasattr(obj, key):
                            setattr(obj, key, value)
                    results["updated"] += 1
                else:
                    results["not_found"] += 1

            session.commit()
        except Exception as e:
            session.rollback()
            results["failed"] += len(chunk)
            results["errors"].append({"chunk": i // chunk_size, "error": str(e)})

    return results


async def bulk_delete(
    session: Session,
    model_class,
    ids: List[Any],
    id_field: str = "euid",
    soft_delete: bool = True,
    chunk_size: int = 100,
) -> Dict[str, Any]:
    """
    Bulk delete objects directly.

    Args:
        session: SQLAlchemy session
        model_class: ORM model class
        ids: List of ID values to delete
        id_field: Field name used for lookup
        soft_delete: Set is_deleted=True vs hard delete
        chunk_size: Items per transaction chunk

    Returns:
        Summary dict with counts
    """
    results = {"deleted": 0, "not_found": 0, "failed": 0, "errors": []}

    for i in range(0, len(ids), chunk_size):
        chunk = ids[i:i + chunk_size]

        try:
            for id_value in chunk:
                obj = session.query(model_class).filter(
                    getattr(model_class, id_field) == id_value
                ).first()

                if obj:
                    if soft_delete and hasattr(obj, 'is_deleted'):
                        obj.is_deleted = True
                    else:
                        session.delete(obj)
                    results["deleted"] += 1
                else:
                    results["not_found"] += 1

            session.commit()
        except Exception as e:
            session.rollback()
            results["failed"] += len(chunk)
            results["errors"].append({"chunk": i // chunk_size, "error": str(e)})

    return results


# Global processor instance
_batch_processor: Optional[BatchProcessor] = None


def get_batch_processor(bdb=None) -> BatchProcessor:
    """
    Get batch processor singleton.

    Args:
        bdb: Optional BLOOMdb3 instance (creates new if not provided)
    """
    global _batch_processor

    if _batch_processor is None or bdb is not None:
        if bdb is None:
            from bloom_lims.db import BLOOMdb3
            bdb = BLOOMdb3()

        settings = get_settings()
        _batch_processor = BatchProcessor(
            bdb,
            chunk_size=getattr(settings.api, 'batch_chunk_size', 100),
            max_concurrent=getattr(settings.api, 'batch_max_concurrent', 5),
        )

    return _batch_processor



"""
BLOOM LIMS API Module

This module provides the versioned API infrastructure for BLOOM LIMS.

API Versioning Strategy:
    - All endpoints are prefixed with /api/v{version}
    - Version 1 is the current stable version
    - Backward compatibility maintained for deprecated endpoints
    - Version negotiation via Accept header (optional)

Usage:
    from bloom_lims.api import create_api_app, api_v1_router

    app = FastAPI()
    app.include_router(api_v1_router)

Rate Limiting:
    from bloom_lims.api import RateLimitMiddleware, rate_limit

    app.add_middleware(RateLimitMiddleware)

    @rate_limit(requests_per_minute=30)
    def my_endpoint():
        pass
"""

from .versioning import (
    APIVersion,
    get_api_version,
    create_versioned_router,
    version_header_dependency,
)
from .v1 import router as api_v1_router
from .rate_limiting import (
    RateLimiter,
    RateLimitMiddleware,
    RateLimitConfig,
    rate_limit,
    get_rate_limiter,
)

__all__ = [
    "APIVersion",
    "get_api_version",
    "create_versioned_router",
    "version_header_dependency",
    "api_v1_router",
    # Rate limiting
    "RateLimiter",
    "RateLimitMiddleware",
    "RateLimitConfig",
    "rate_limit",
    "get_rate_limiter",
]




"""Direct API coverage for the execution queue router."""

from __future__ import annotations

import os
import secrets
import sys

import pytest
from fastapi.testclient import TestClient

from bloom_lims.api.v1.dependencies import APIUser, require_api_auth

os.environ["BLOOM_DEV_AUTH_BYPASS"] = "true"
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from main import app  # noqa: E402


@pytest.fixture(autouse=True)
def _clear_overrides():
    yield
    app.dependency_overrides.clear()


def _opaque(prefix: str) -> str:
    return f"{prefix}-{secrets.token_hex(8)}"


def _internal_rw_user() -> APIUser:
    return APIUser(
        email="execution-api@example.com",
        user_id=_opaque("user"),
        roles=["INTERNAL_READ_WRITE"],
        auth_source="token",
        token_scope="internal_rw",
        token_id=_opaque("token"),
    )


def _internal_ro_user() -> APIUser:
    return APIUser(
        email="execution-api-ro@example.com",
        user_id=_opaque("user"),
        roles=["INTERNAL_READ_ONLY"],
        auth_source="token",
        token_scope="internal_ro",
        token_id=_opaque("token"),
    )


def test_execution_queue_router_lists_queues_and_registers_worker():
    app.dependency_overrides[require_api_auth] = _internal_rw_user
    worker_key = f"worker://pytest/{_opaque('worker')}"

    with TestClient(app) as client:
        queues = client.get("/api/v1/execution/queues")
        assert queues.status_code == 200, queues.text
        queue_keys = {item["queue_key"] for item in queues.json()}
        assert "extraction_prod" in queue_keys

        registered = client.post(
            "/api/v1/execution/actions/register-worker",
            json={
                "worker_key": worker_key,
                "display_name": "Pytest Execution Worker",
                "worker_type": "SERVICE",
                "status": "ONLINE",
                "capabilities": ["wetlab.extraction"],
                "max_concurrent_leases": 2,
                "heartbeat_ttl_seconds": 120,
            },
        )
        assert registered.status_code == 200, registered.text
        worker = registered.json()
        assert worker["worker_key"] == worker_key
        assert worker["capabilities"] == ["wetlab.extraction"]
        assert worker["max_concurrent_leases"] == 2

        workers = client.get("/api/v1/execution/workers")
        assert workers.status_code == 200, workers.text
        worker_keys = {item["worker_key"] for item in workers.json()}
        assert worker_key in worker_keys


def test_execution_queue_router_register_worker_requires_write_permission():
    app.dependency_overrides[require_api_auth] = _internal_ro_user

    with TestClient(app) as client:
        response = client.post(
            "/api/v1/execution/actions/register-worker",
            json={
                "worker_key": f"worker://pytest/{_opaque('worker')}",
                "display_name": "Read Only Worker",
                "worker_type": "SERVICE",
                "status": "ONLINE",
            },
        )

    assert response.status_code == 403
    assert "Write permission required" in response.text


def test_execution_queue_router_maps_missing_queue_to_404():
    app.dependency_overrides[require_api_auth] = _internal_rw_user

    with TestClient(app) as client:
        response = client.get(f"/api/v1/execution/queues/{_opaque('missing-queue')}")

    assert response.status_code == 404
    assert "Queue not found" in response.text



"""TapDB-backed repository for BLOOM unified search."""

from __future__ import annotations

import json
from collections import defaultdict
from datetime import datetime
from typing import Any, Dict, Iterable, List, Tuple

from sqlalchemy import and_, cast, func, or_
from sqlalchemy.sql.sqltypes import String

from bloom_lims.db import BLOOMdb3
from bloom_lims.search.contracts import SearchJSONFilter, SearchRequest, SearchResultItem


class SearchRepository:
    """Repository that executes unified search queries using BLOOMdb3/TapDB models."""

    def __init__(self, bdb: BLOOMdb3):
        self._bdb = bdb
        self._session = bdb.session
        self._classes = bdb.Base.classes

    def search(self, request: SearchRequest) -> Tuple[List[SearchResultItem], Dict[str, Dict[str, int]], bool]:
        """Run search for requested record types and return normalized items + facets."""
        handlers = {
            "instance": self._search_instances,
            "template": self._search_templates,
            "lineage": self._search_lineages,
            "audit": self._search_audit,
        }

        all_items: List[SearchResultItem] = []
        record_type_counts: Dict[str, int] = {}
        category_counts: Dict[str, int] = defaultdict(int)
        truncated = False

        for record_type in request.record_types:
            handler = handlers.get(record_type)
            if handler is None:
                continue
            items, total_count, is_truncated = handler(request)
            all_items.extend(items)
            record_type_counts[record_type] = total_count
            truncated = truncated or is_truncated
            for item in items:
                if item.category:
                    category_counts[item.category] += 1

        facets = {
            "record_type": record_type_counts,
            "category": dict(sorted(category_counts.items())),
        }
        return all_items, facets, truncated

    def _search_instances(self, request: SearchRequest) -> Tuple[List[SearchResultItem], int, bool]:
        model = self._classes.generic_instance
        query = self._session.query(model)

        query = self._apply_common_filters(query, model, request, include_type_filters=True)
        query = self._apply_query_text_filter(
            query,
            request.query,
            [
                model.euid,
                model.name,
                model.category,
                model.type,
                model.subtype,
                cast(model.json_addl, String),
            ],
        )
        query = self._apply_json_filters(query, model.json_addl, request)
        query = self._apply_query_sort(query, model, request, timestamp_field="created_dt")

        return self._materialize_items(
            query=query,
            record_type="instance",
            request=request,
            to_item=lambda row: SearchResultItem(
                record_type="instance",
                euid=str(getattr(row, "euid", "") or ""),
                name=str(getattr(row, "name", "") or ""),
                category=str(getattr(row, "category", "") or ""),
                type=str(getattr(row, "type", "") or ""),
                subtype=str(getattr(row, "subtype", "") or ""),
                status=str(getattr(row, "bstatus", "") or ""),
                created_dt=getattr(row, "created_dt", None),
                modified_dt=getattr(row, "modified_dt", None),
                timestamp=getattr(row, "created_dt", None) or getattr(row, "modified_dt", None),
                metadata={"json_addl": getattr(row, "json_addl", {}) or {}},
            ),
        )

    def _search_templates(self, request: SearchRequest) -> Tuple[List[SearchResultItem], int, bool]:
        model = self._classes.generic_template
        query = self._session.query(model)

        query = self._apply_common_filters(query, model, request, include_type_filters=True)
        query = self._apply_query_text_filter(
            query,
            request.query,
            [
                model.euid,
                model.name,
                model.category,
                model.type,
                model.subtype,
                cast(model.json_addl, String),
            ],
        )
        query = self._apply_json_filters(query, model.json_addl, request)
        query = self._apply_query_sort(query, model, request, timestamp_field="created_dt")

        return self._materialize_items(
            query=query,
            record_type="template",
            request=request,
            to_item=lambda row: SearchResultItem(
                record_type="template",
                euid=str(getattr(row, "euid", "") or ""),
                name=str(getattr(row, "name", "") or ""),
                category=str(getattr(row, "category", "") or ""),
                type=str(getattr(row, "type", "") or ""),
                subtype=str(getattr(row, "subtype", "") or ""),
                status=str(getattr(row, "bstatus", "") or ""),
                created_dt=getattr(row, "created_dt", None),
                modified_dt=getattr(row, "modified_dt", None),
                timestamp=getattr(row, "created_dt", None) or getattr(row, "modified_dt", None),
                metadata={
                    "json_addl": getattr(row, "json_addl", {}) or {},
                    "json_addl_schema": getattr(row, "json_addl_schema", None),
                    "instance_prefix": getattr(row, "instance_prefix", ""),
                },
            ),
        )

    def _search_lineages(self, request: SearchRequest) -> Tuple[List[SearchResultItem], int, bool]:
        model = self._classes.generic_instance_lineage
        query = self._session.query(model)

        query = self._apply_common_filters(query, model, request, include_type_filters=False)
        query = self._apply_query_text_filter(
            query,
            request.query,
            [
                model.euid,
                model.name,
                model.category,
                model.parent_type,
                model.child_type,
                model.relationship_type,
                cast(model.json_addl, String),
            ],
        )
        query = self._apply_json_filters(query, model.json_addl, request)
        query = self._apply_query_sort(query, model, request, timestamp_field="created_dt")

        return self._materialize_items(
            query=query,
            record_type="lineage",
            request=request,
            to_item=lambda row: SearchResultItem(
                record_type="lineage",
                euid=str(getattr(row, "euid", "") or ""),
                name=str(getattr(row, "name", "") or ""),
                category=str(getattr(row, "category", "") or ""),
                type=str(getattr(row, "parent_type", "") or ""),
                subtype=str(getattr(row, "child_type", "") or ""),
                status=str(getattr(row, "relationship_type", "") or ""),
                created_dt=getattr(row, "created_dt", None),
                modified_dt=getattr(row, "modified_dt", None),
                timestamp=getattr(row, "created_dt", None) or getattr(row, "modified_dt", None),
                metadata={
                    "parent_instance_euid": getattr(getattr(row, "parent_instance", None), "euid", None),
                    "child_instance_euid": getattr(getattr(row, "child_instance", None), "euid", None),
                    "relationship_type": getattr(row, "relationship_type", ""),
                    "json_addl": getattr(row, "json_addl", {}) or {},
                },
            ),
        )

    def _search_audit(self, request: SearchRequest) -> Tuple[List[SearchResultItem], int, bool]:
        model = self._classes.audit_log
        query = self._session.query(model)

        if not request.include_deleted and hasattr(model, "is_deleted"):
            query = query.filter(model.is_deleted.is_(False))

        if request.categories and hasattr(model, "category"):
            query = query.filter(func.lower(model.category).in_([c.lower() for c in request.categories]))

        if request.statuses:
            query = query.filter(func.lower(model.operation_type).in_([s.lower() for s in request.statuses]))

        if request.type_names:
            query = query.filter(func.lower(model.rel_table_name).in_([t.lower() for t in request.type_names]))

        if request.created_dt_start:
            query = query.filter(model.changed_at >= request.created_dt_start)
        if request.created_dt_end:
            query = query.filter(model.changed_at <= request.created_dt_end)

        query = self._apply_query_text_filter(
            query,
            request.query,
            [
                model.euid,
                model.rel_table_name,
                model.column_name,
                model.rel_table_euid_fk,
                model.changed_by,
                model.operation_type,
                model.old_value,
                model.new_value,
                cast(model.json_addl, String),
            ],
        )

        query = self._apply_json_filters(query, model.json_addl, request)
        query = self._apply_query_sort(query, model, request, timestamp_field="changed_at")

        return self._materialize_items(
            query=query,
            record_type="audit",
            request=request,
            to_item=lambda row: SearchResultItem(
                record_type="audit",
                euid=str(getattr(row, "euid", "") or ""),
                name=str(getattr(row, "rel_table_name", "") or ""),
                category=str(getattr(row, "category", "") or ""),
                type=str(getattr(row, "rel_table_name", "") or ""),
                subtype=str(getattr(row, "column_name", "") or ""),
                status=str(getattr(row, "operation_type", "") or ""),
                created_dt=getattr(row, "changed_at", None),
                modified_dt=getattr(row, "changed_at", None),
                timestamp=getattr(row, "changed_at", None),
                metadata={
                    "rel_table_name": getattr(row, "rel_table_name", ""),
                    "column_name": getattr(row, "column_name", ""),
                    "rel_table_euid_fk": getattr(row, "rel_table_euid_fk", ""),
                    "changed_by": getattr(row, "changed_by", ""),
                    "old_value": getattr(row, "old_value", ""),
                    "new_value": getattr(row, "new_value", ""),
                    "json_addl": getattr(row, "json_addl", {}) or {},
                },
            ),
        )

    def _materialize_items(self, *, query, record_type: str, request: SearchRequest, to_item):
        total_count = query.count()
        rows = query.limit(request.max_scan).all()
        truncated = total_count > len(rows)
        items = [to_item(row) for row in rows]
        return items, total_count, truncated

    def _apply_common_filters(self, query, model, request: SearchRequest, include_type_filters: bool):
        if not request.include_deleted and hasattr(model, "is_deleted"):
            query = query.filter(model.is_deleted.is_(False))

        if request.categories and hasattr(model, "category"):
            query = query.filter(func.lower(model.category).in_([c.lower() for c in request.categories]))

        if include_type_filters and request.type_names and hasattr(model, "type"):
            query = query.filter(func.lower(model.type).in_([t.lower() for t in request.type_names]))

        if include_type_filters and request.subtype_names and hasattr(model, "subtype"):
            query = query.filter(func.lower(model.subtype).in_([s.lower() for s in request.subtype_names]))

        if request.statuses and hasattr(model, "bstatus"):
            query = query.filter(func.lower(model.bstatus).in_([s.lower() for s in request.statuses]))

        if request.created_dt_start and hasattr(model, "created_dt"):
            query = query.filter(model.created_dt >= request.created_dt_start)
        if request.created_dt_end and hasattr(model, "created_dt"):
            query = query.filter(model.created_dt <= request.created_dt_end)

        return query

    def _apply_query_text_filter(self, query, search_text: str, columns: Iterable[Any]):
        if not search_text:
            return query

        pattern = f"%{search_text}%"
        conditions = [column.ilike(pattern) for column in columns if column is not None]
        if conditions:
            query = query.filter(or_(*conditions))
        return query

    def _apply_json_filters(self, query, json_column, request: SearchRequest):
        if not request.json_filters:
            return query

        conditions = []
        for json_filter in request.json_filters:
            condition = self._build_json_filter_condition(json_column, json_filter)
            if condition is not None:
                conditions.append(condition)

        if not conditions:
            return query

        if request.search_mode == "all":
            return query.filter(and_(*conditions))
        return query.filter(or_(*conditions))

    def _build_json_filter_condition(self, json_column, json_filter: SearchJSONFilter):
        path_parts = [part for part in json_filter.path.split(".") if part]
        extracted = func.jsonb_extract_path_text(json_column, *path_parts)

        if json_filter.op == "exists":
            return extracted.isnot(None)

        if json_filter.op == "between":
            start = json_filter.start
            end = json_filter.end
            if start is None and end is None:
                return None
            conditions = [extracted.isnot(None), extracted != ""]
            if start is not None:
                conditions.append(extracted >= start.strftime("%Y-%m-%d"))
            if end is not None:
                conditions.append(extracted <= end.strftime("%Y-%m-%d"))
            return and_(*conditions)

        if json_filter.op == "contains":
            if path_parts:
                payload = self._build_nested_payload(path_parts, json_filter.value)
            else:
                payload = json_filter.value
            return json_column.op("@>")(json.dumps(payload, default=str))

        if json_filter.op == "in":
            values = json_filter.values or ([] if json_filter.value is None else [json_filter.value])
            if not values:
                return None
            return or_(*[extracted == str(value) for value in values])

        if json_filter.op == "ilike":
            if json_filter.value is None:
                return None
            return extracted.ilike(f"%{json_filter.value}%")

        if json_filter.value is None:
            return extracted.is_(None)
        return extracted == str(json_filter.value)

    def _build_nested_payload(self, path_parts: List[str], value: Any) -> Dict[str, Any]:
        payload: Any = value
        for key in reversed(path_parts):
            payload = {key: payload}
        return payload

    def _apply_query_sort(self, query, model, request: SearchRequest, timestamp_field: str):
        if request.sort_by == "euid" and hasattr(model, "euid"):
            order_column = model.euid
        elif request.sort_by == "name" and hasattr(model, "name"):
            order_column = model.name
        else:
            order_column = getattr(model, timestamp_field, None) or getattr(model, "created_dt", None)
            if order_column is None and hasattr(model, "euid"):
                order_column = model.euid

        if request.sort_order == "asc":
            return query.order_by(order_column.asc())
        return query.order_by(order_column.desc())



#!/usr/bin/env python3

import pytest

from bloom_lims.db import BLOOMdb3
from bloom_lims.bobjs import BloomWorkflow, BloomWorkflowStep, BloomObj, BloomReagent
import sys
import os

bob_wf = BloomWorkflow(BLOOMdb3(),cfg_fedex=True)

bob_wfs = BloomWorkflowStep(BLOOMdb3(), cfg_fedex=True)

bob_rg = BloomReagent(BLOOMdb3(), cfg_fedex=True)

idt_plate_euid = bob_rg.create_rgnt_24w_plate_TEST('idt-probes-rare-mendelian')
naoh_plate_euid = bob_rg.create_rgnt_24w_plate_TEST('naoh')

from random import randint

ASSAY = "workflow/assay/hla-typing/1.2" if int(sys.argv[2]) > 0 else "workflow/assay/carrier-screen/3.9"

giab_cx, giab_mx = bob_wf.create_container_with_content(
        ("container", "tube", "tube-generic-10ml", "1.0"),
        ("content", "control", "giab-HG002", "1.0")
    )   
 
cfsynctl_cx, cfsynctl_mx = bob_wf.create_container_with_content(
        ("container", "tube", "tube-10ml-glass", "1.0"),
        ("content", "control", "synthetic-cfdna", "1.0")
    )  

ntc_cx, ntc_mx = bob_wf.create_container_with_content( 
        ("container", "tube", "tube-eppi-1.5ml", "1.0"),
        ("content", "control", "water-ntc", "1.0")
    )

## This needs to be done better
CA = [{"cont_address": {
                  "name": "A1",
                  "row_name": "A",
                  "col_name": "1",
                  "row_idx": "0",
                  "col_idx": "0"
                }},{"cont_address": {
                  "name": "A2",
                  "row_name": "A",
                  "col_name": "2",
                  "row_idx": "1",
                  "col_idx": "0"
                }},{"cont_address": {
                  "name": "A3",
                  "row_name": "A",
                  "col_name": "3",
                  "row_idx": "0",
                  "col_idx": "2"
                }}]
                
rgnt_plate_glob = bob_wf.create_instance_by_template_components(
    "container", "plate", "fixed-plate-24", "1.0"
)
rgnt_plate = rgnt_plate_glob[0][0]
rgnt_plate_wells = rgnt_plate_glob[1]
for i in rgnt_plate_wells:
    bob_wf.create_generic_instance_lineage_by_euids(i.euid, bob_wf.create_instance_by_template_components("content","reagent","naoh","1.0")[0][0].euid)


      
new_rack = bob_wf.create_instance_by_template_components(
    "container", "rack", "tube-rack-4-empty", "1.0"
)[0][0]

bob_wf.create_generic_instance_lineage_by_euids(new_rack.euid, ntc_cx.euid)
ntc_cx.json_addl['cont_address']=CA[0]['cont_address']

bob_wf.create_generic_instance_lineage_by_euids(new_rack.euid, cfsynctl_cx.euid)
cfsynctl_cx.json_addl['cont_address']=CA[1]['cont_address']

bob_wf.create_generic_instance_lineage_by_euids(new_rack.euid, giab_cx.euid)
giab_cx.json_addl['cont_address']=CA[2]['cont_address']


def set_status(b, obj, status):
    action = "action/core/set_object_status/1.0"
    action_group = "core"

    action_ds = obj.json_addl["action_groups"][action_group]["actions"][action]
    action_ds["captured_data"]["object_status"] = status
    b.do_action(obj.euid, action, action_group, action_ds)

TUBES=[]

def create_tubes(n=1):
    ctr = 0
    while ctr < n:

        #records = (
        #    bob_wfs.session.query(bob_wfs.Base.classes.workflow_template)
        #    .filter(
        #        bob_wfs.Base.classes.workflow_template.subtype
        #        == "accession-package-kit-tubes-testreq"
        #    )
        #    .all()
        #)

        #wf = bob_wfs.create_instances(records[0].euid)[0][0]

        # Clinical ACC Queue
        #    accessioning-RnD
         
        #from IPython import embed; embed()
        #raise
        # 
        # 
        wf = bob_wf.query_instance_by_component_v2("workflow", "assay", "accessioning-RnD", "1.0")[0]

        action_group = "accessioning"
        action = "action/accessioning-ay/create_package_and_first_workflow_step_assay_root/1.0"
        action_data = wf.json_addl["action_groups"][action_group]["actions"][action]
        # action_data = wf.json_addl["actions"]["create_package_and_first_workflow_step"]
        action_data["captured_data"]["Tracking Number"] = "1001897582860000245100773464327825"
        action_data["captured_data"]["Fedex Tracking Data"] = {}


        wfs=bob_wf.do_action(wf.euid, action, action_group, action_data)
        set_status(bob_wfs, wfs, "in_progress")

        #wfs = bob_wf.do_action_create_package_and_first_workflow_step_assay(wf.euid, action_data)
    
        assert hasattr(wfs, "euid") == True

        b_action_group = "create_child"  # change to child
        #create_package_and_first_workflow_step_assay
        b_action = "action/workflow_step_accessioning/create_child_container_and_link_child_workflow_step/1.0" # "action/workflow_step_accessioning/create_child_container_and_link_child_workflow_step/1.0"
        wfs_action_data = wfs.json_addl["action_groups"][b_action_group]["actions"][b_action]

        wfs_action_data["captured_data"]["Tracking Number"] = "1001897582860000245100773464327825"
        wfs_action_data["captured_data"]["Fedex Tracking Data"] = {}
        bob_wfs.do_action(wfs.euid, b_action, b_action_group, wfs_action_data)

        
        child_wfs = ""
        for i in wfs.parent_of_lineages:
            if i.child_instance.type == "accessioning-steps":
                child_wfs = i.child_instance

        assert hasattr(child_wfs, "euid") == True
        c_action_group = "specimen_actions"  # change to child
        c_action = "action/workflow_step_accessioning/create_child_container_and_link_child_workflow_step_specimen/1.0"
        c_wfs_action_data = child_wfs.json_addl["action_groups"][c_action_group]["actions"][
            c_action
        ]

        bob_wfs.do_action(child_wfs.euid, c_action, c_action_group, c_wfs_action_data)
        set_status(bob_wfs, wfs, "complete")

        set_status(bob_wfs, child_wfs, "in_progress")        

        new_child_wfs = ""
        for i in child_wfs.parent_of_lineages:
            if i.child_instance.category == "workflow_step":
                new_child_wfs = i.child_instance
        assert hasattr(new_child_wfs, "euid") == True




        trf_wfs = bob_wfs.do_action(
            new_child_wfs.euid,
            "action/workflow_step_accessioning/create_test_req_and_link_child_workflow_step_dup/1.0",
            "test_req",
            new_child_wfs.json_addl["action_groups"]["test_req"]["actions"][
                "action/workflow_step_accessioning/create_test_req_and_link_child_workflow_step_dup/1.0"
            ],
        )
        set_status(bob_wfs, new_child_wfs, "in_progress")
        set_status(bob_wfs, child_wfs, "complete")


        assert hasattr(trf_wfs, "euid") == True


        trf_child_wfs = ""
        trf_child_cont = ""
        for i in trf_wfs.parent_of_lineages:
            if i.child_instance.category == "workflow_step":
                trf_child_wfs = i.child_instance
            if i.child_instance.category == "container":
                trf_child_cont = i.child_instance

        trf = ""
        for i in trf_child_cont.child_of_lineages:
            if i.parent_instance.category == "test_requisition":
                trf = i.parent_instance


        trf_assay_data = trf.json_addl["action_groups"]["test_requisitions"]["actions"][
            "action/test_requisitions/add_container_to_assay_q/1.0"
        ]
        # This is super brittle, how I am currently linking Assay to TestReq...
        # = tr.json_addl["actions"]["add_container_to_assay_q"]
        trf_assay_data["captured_data"]["assay_selection"] = ASSAY
        trf_assay_data["captured_data"]["Container EUID"] = trf_child_cont.euid

        wfs_queue = bob_wfs.do_action(
            trf.euid,
            action_group="test_requisitions",
            action="action/test_requisitions/add_container_to_assay_q/1.0",
            action_ds=trf_assay_data,
        )
        assert hasattr(wfs_queue, "euid") == True
        set_status(bob_wfs, wfs_queue, "in_progress")


        scanned_bcs = trf_child_cont.euid

        ctr = ctr + 1
        print('CCCCCCCCCCC',ctr)
        TUBES.append(trf_child_cont.euid)



        wset_q = wfs
        wset_q_axn = "action/move-queues/move-among-ay-top-queues/1.0"
        wset_q_axn_grp = "acc-queue-move"
        wset_q_ad = wset_q.json_addl["action_groups"][wset_q_axn_grp][
            "actions"][wset_q_axn]  
        wset_q_ad["captured_data"]["q_selection"] = "workflow_step/queue/plasma-isolation-queue-exception/1.0" if randint(0,13) > 10 else "workflow_step/queue/plasma-isolation-queue-removed/1.0"
        bob_wfs.do_action(
            wset_q.euid,
            action_group=wset_q_axn_grp,
            action=wset_q_axn,
            action_ds=wset_q_ad
        )


def fill_plates(tubes=[]):
        # Create some controls to add to the plate!
        giab_cx, giab_mx = bob_wf.create_container_with_content(
            ("container", "tube", "tube-generic-10ml", "1.0"),
            ("content", "control", "giab-HG002", "1.0")
        )   
 
        cfsynctl_cx, cfsynctl_mx = bob_wf.create_container_with_content(
            ("container", "tube", "tube-10ml-glass", "1.0"),
            ("content", "control", "synthetic-cfdna", "1.0")
        )  

        ntc_cx, ntc_mx = bob_wf.create_container_with_content( 
            ("container", "tube", "tube-eppi-1.5ml", "1.0"),
            ("content", "control", "water-ntc", "1.0")
        )
        trf_child_cont = bob_wf.get_by_euid(tubes[-1])
        tubes.append(giab_cx.euid)
        tubes.append(cfsynctl_cx.euid)
        tubes.append(ntc_cx.euid)

        scanned_bcs = "\n".join(tubes)
        q_wfs = ""
        for i in trf_child_cont.child_of_lineages:
            if i.parent_instance.subtype == "plasma-isolation-queue-available":
                q_wfs = i.parent_instance

        piso_q_action_data = q_wfs.json_addl["action_groups"]["tube_xfer"]["actions"][
            "action/workflow_step_queue/link_tubes_auto/1.0"
        ]
        piso_q_action_data["captured_data"]["discard_barcodes"] = scanned_bcs
        wfs_plasma = bob_wfs.do_action(
            q_wfs.euid,
            action_group="tube_xfer",
            action="action/workflow_step_queue/link_tubes_auto/1.0",
            action_ds=piso_q_action_data,
        )  # _link_tubes_auto(wfs_queue.euid, piso_q_action_data)

        plasma_cont = None
        for i in trf_child_cont.parent_of_lineages:
            if i.child_instance.subtype == "tube-generic-10ml":
                plasma_cont = i.child_instance
        scanned_bcs_plasma = plasma_cont.euid

        for i in plasma_cont.child_of_lineages:
            if i.parent_instance.category == "workflow_step":
                pi_wfs = i.parent_instance
        
        wfset_wf = pi_wfs.child_of_lineages[0].parent_instance
        
        action_ds_plasma = pi_wfs.json_addl["action_groups"]["fill_plate"]["actions"][
            "action/workflow_step_queue/fill_plate_undirected/1.0"
        ]
        # wfs_plasma.json_addl["actions"]["fill_plate_undirected"]
        action_ds_plasma["captured_data"]["discard_barcodes"] = scanned_bcs_plasma
        wfs_plt = bob_wfs.do_action(
            pi_wfs.euid,
            action_group="fill_plate",
            action="action/workflow_step_queue/fill_plate_undirected/1.0",
            action_ds=action_ds_plasma,
        )
        set_status(bob_wfs, pi_wfs, "complete")

        ### ENDING WITH ANEXTRACTION PLATE!  Need to check quant and add more from here.

        plt_fill_wfs = ""
        for i in wfs_plt.parent_of_lineages:
            if i.child_instance.category == "workflow_step":
                plt_fill_wfs = i.child_instance

        action_data_dat = plt_fill_wfs.json_addl["action_groups"]["plate_operations"][
            "actions"
        ]["action/workflow_step_plate_operations/cfdna_quant/1.0"]
        # action_data_dat = wfs_plt.json_addl["actions"]["cfdna_quant"]
        action_data_dat["captured_data"]["gdna_quant"] = ""
        yy = bob_wfs.do_action(
            plt_fill_wfs.euid,
            action_group="plate_operations",
            action="action/workflow_step_plate_operations/cfdna_quant/1.0",
            action_ds=action_data_dat,
        )
        set_status(bob_wfs, plt_fill_wfs, "complete")

        eplt = None
        for i in plt_fill_wfs.parent_of_lineages:
            if i.child_instance.category == "workflow_step":
                eplt = i.child_instance
        set_status(bob_wfs, eplt, "in_progress")

        next_plate = ""
        for i in plt_fill_wfs.parent_of_lineages:
            if i.child_instance.type == "plate":
                next_plate = i.child_instance.euid

        stamp_action_data = plt_fill_wfs.json_addl["action_groups"]["plate_operations"][
            "actions"
        ]["action/workflow_step_plate_operations/stamp_copy_plate/1.0"]
        stamp_action_data["captured_data"]["plate_euid"] = next_plate

        xx = bob_wfs.do_action(
            plt_fill_wfs.euid,
            action_group="plate_operations",
            action="action/workflow_step_plate_operations/stamp_copy_plate/1.0",
            action_ds=stamp_action_data,
        )

        for i in plt_fill_wfs.parent_of_lineages:
            if i.child_instance.type == "plate-operations":
                sec_stamp_wfs = i.child_instance

        for i in sec_stamp_wfs.parent_of_lineages:
            if i.child_instance.category == "container":
                next_plate2 = i.child_instance.euid

        stamp_action_data2 = sec_stamp_wfs.json_addl["action_groups"]["plate_operations"][
            "actions"
        ]["action/workflow_step_plate_operations/stamp_copy_plate/1.0"]
        stamp_action_data2["captured_data"]["plate_euid"] = next_plate2
        stamp_wfs2 = bob_wfs.do_action(
            sec_stamp_wfs.euid,
            action_group="plate_operations",
            action="action/workflow_step_plate_operations/stamp_copy_plate/1.0",
            action_ds=stamp_action_data2,
        )


        # make a control
        rgnt = bob_wfs.create_instance_by_template_components(
            "content", "control", "giab-HG002", "1.0"
        )[0][0]
        # put in a tube
        tube = bob_wfs.create_instance_by_template_components(
            "container", "tube", "tube-generic-10ml", "1.0"
        )[0][0]
        # put the reagent in the tube

        bob_wfs.create_generic_instance_lineage_by_euids(tube.euid, rgnt.euid)


        wset_q = wfset_wf
        wset_q_axn = "action/move-queues/move-among-ay-top-queues/1.0"
        wset_q_axn_grp = "queue-move"
        wset_q_ad = wset_q.json_addl["action_groups"][wset_q_axn_grp][
            "actions"][wset_q_axn]  
        wset_q_ad["captured_data"]["q_selection"] = "workflow_step/queue/plasma-isolation-queue-exception/1.0" if randint(0,5) > 4 else "workflow_step/queue/plasma-isolation-queue-removed/1.0"
        bob_wfs.do_action(
            wset_q.euid,
            action_group=wset_q_axn_grp,
            action=wset_q_axn,
            action_ds=wset_q_ad
        )
        

n_tubes = 20 if int(sys.argv[1]) > 20 else int(sys.argv[1])

create_tubes(n_tubes)

fill_plates(tubes=TUBES)



"""
Route coverage gap tests (GUI routes).

These are intentionally minimal, but they must execute the handler body for
routes that are otherwise easy to miss (because 422s don't exercise code).
"""

from __future__ import annotations

import json
import os
import sys
from pathlib import Path
from unittest.mock import AsyncMock, patch

import pytest

# Disable OAuth for GUI tests and use dev API auth bypass.
os.environ["BLOOM_OAUTH"] = "no"
os.environ["BLOOM_DEV_AUTH_BYPASS"] = "true"

from fastapi.responses import RedirectResponse
from fastapi.testclient import TestClient

sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
import main


@pytest.fixture
def client() -> TestClient:
    return TestClient(main.app, raise_server_exceptions=False)


def _warm_session(client: TestClient) -> None:
    # Any route with Depends(require_auth) will populate session when BLOOM_OAUTH=no.
    resp = client.get("/lims")
    assert resp.status_code == 200


def _get_any_template_euid(bdb) -> str:
    GT = bdb.Base.classes.generic_template
    row = bdb.session.query(GT).filter(GT.is_deleted == False).first()
    assert row is not None
    return row.euid


def _create_plate_instance(client: TestClient) -> dict:
    resp = client.post(
        "/api/v1/object-creation/create",
        json={
            "category": "container",
            "type": "plate",
            "subtype": "fixed-plate-96",
            "version": "1.0",
            "name": "coverage-plate",
        },
    )
    assert resp.status_code == 200, resp.text
    return resp.json()


def _create_queue_instance(client: TestClient) -> dict:
    resp = client.post(
        "/api/v1/object-creation/create",
        json={
            "category": "workflow_step",
            "type": "queue",
            "subtype": "all-purpose",
            "version": "1.0",
            "name": "coverage-queue",
        },
    )
    assert resp.status_code == 200, resp.text
    return resp.json()


def test_auth_callback_aliases(client: TestClient) -> None:
    resp = client.get("/auth/callback")
    assert resp.status_code == 200
    assert "Processing" in resp.text or "login" in resp.text.lower()

    with patch(
        "bloom_lims.gui.routes.auth._complete_cognito_login",
        new=AsyncMock(return_value=RedirectResponse(url="/", status_code=303)),
    ):
        resp = client.post("/auth/callback", json={"access_token": "x", "id_token": "y"}, follow_redirects=False)
        assert resp.status_code == 303


def test_list_scripts_executes_handler(client: TestClient) -> None:
    resp = client.get("/list-scripts", params={"directory": str(Path(__file__).resolve().parents[1] / "static")})
    assert resp.status_code == 200, resp.text
    assert "scripts" in resp.json()


def test_legacy_graph_alias_routes_are_removed(client: TestClient) -> None:
    resp = client.get("/dag", params={"start_euid": "CX-1"}, follow_redirects=False)
    assert resp.status_code == 404

    resp = client.get("/dag_explorer", follow_redirects=False)
    assert resp.status_code == 404

    resp = client.get("/graph", params={"start_euid": "CX-1"}, follow_redirects=False)
    assert resp.status_code == 404


def test_cogs_and_node_info_routes(client: TestClient) -> None:
    _warm_session(client)
    created = client.post(
        "/api/v1/object-creation/create",
        json={"category": "container", "type": "tube", "subtype": "tube-generic-10ml", "version": "1.0", "name": "cogs-tube"},
    ).json()
    euid = created["euid"]

    resp = client.get("/calculate_cogs_children", params={"euid": euid})
    assert resp.status_code == 200

    resp = client.get("/calculate_cogs_parents", params={"euid": euid})
    assert resp.status_code == 200

    resp = client.get("/get_node_info", params={"euid": euid})
    assert resp.status_code == 200
    assert resp.json().get("euid") == euid

    resp = client.get("/get_node_property", params={"euid": euid, "key": "name"})
    assert resp.status_code in (200, 404, 500), resp.text


def test_plate_routes_with_and_without_valid_plate(client: TestClient) -> None:
    _warm_session(client)

    # plate_carosel2 is retired from modern GUI.
    resp = client.get("/plate_carosel2", params={"plate_euid": "CX-NOT-REAL"})
    assert resp.status_code == 404

    # plate_visualization: create a real plate with wells
    plate = _create_plate_instance(client)
    resp = client.get("/plate_visualization", params={"plate_euid": plate["euid"]})
    assert resp.status_code == 200
    assert "text/html" in resp.headers.get("content-type", "")

    # get_related_plates is retired from modern GUI.
    resp = client.get("/get_related_plates", params={"main_plate": plate["euid"]})
    assert resp.status_code == 404


def test_queue_details_renders(client: TestClient) -> None:
    _warm_session(client)
    queue = _create_queue_instance(client)

    resp = client.get("/queue_details", params={"queue_euid": queue["euid"], "page": 1})
    assert resp.status_code == 200
    assert "text/html" in resp.headers.get("content-type", "")


def test_legacy_uuid_alias_routes_are_removed(client: TestClient) -> None:
    _warm_session(client)
    details_resp = client.get("/uuid_details", params={"euid": "CX-REMOVED"}, follow_redirects=False)
    assert details_resp.status_code == 404

    restore_resp = client.get("/un_delete_by_uuid", params={"euid": "CX-REMOVED"}, follow_redirects=False)
    assert restore_resp.status_code == 404

    uuid_query_resp = client.get("/uuid_details", params={"uuid": "00000000-0000-0000-0000-000000000000"})
    assert uuid_query_resp.status_code == 404


def test_euid_details_page_does_not_emit_uuid_alias_links(client: TestClient) -> None:
    _warm_session(client)
    obj = client.post(
        "/api/v1/object-creation/create",
        json={"category": "container", "type": "tube", "subtype": "tube-generic-10ml", "version": "1.0", "name": "alias-guard"},
    ).json()

    resp = client.get("/euid_details", params={"euid": obj["euid"]})
    assert resp.status_code == 200
    assert "/uuid_details?uuid=" not in resp.text
    assert f"/dindex2?start_euid={obj['euid']}" in resp.text


def test_user_audit_logs_renders(client: TestClient) -> None:
    _warm_session(client)
    resp = client.get("/user_audit_logs", params={"username": "tester@example.com"})
    assert resp.status_code == 200
    assert "text/html" in resp.headers.get("content-type", "")


def test_create_instance_form_renders(client: TestClient, bdb) -> None:
    _warm_session(client)
    template_euid = _get_any_template_euid(bdb)
    resp = client.get(f"/create_instance/{template_euid}")
    assert resp.status_code == 410


def test_file_set_urls_and_admin_template_pages(client: TestClient, bdb) -> None:
    _warm_session(client)
    # Legacy file set GUI routes are removed in Dewey hard-cut mode.
    resp = client.get("/file_set_urls", params={"fs_euid": "FS-NOT-REAL"})
    assert resp.status_code == 404

    template_euid = _get_any_template_euid(bdb)
    # Admin template editor was retired.
    resp = client.get("/admin_template", params={"euid": template_euid})
    assert resp.status_code == 410


def test_query_by_euids_create_and_download_file_flows(client: TestClient) -> None:
    _warm_session(client)
    # Legacy file GUI handlers are removed in Dewey hard-cut mode.
    resp = client.post("/query_by_euids", data={"file_euids": "FX-1"})
    assert resp.status_code == 200
    assert "text/html" in resp.headers.get("content-type", "")

    resp = client.post(
        "/create_file",
        data={"name": "upload-1"},
        files=[
            ("file_data", ("upload.txt", b"hello", "text/plain")),
            ("directory", (".ignored", b"", "application/octet-stream")),
        ],
    )
    assert resp.status_code == 404

    resp = client.post(
        "/download_file",
        data={
            "euid": "FX-1",
            "download_type": "flat",
            "create_metadata_file": "no",
            "ret_json": "1",
        },
    )
    assert resp.status_code == 404

    resp = client.get("/delete_temp_file", params={"filename": "downloaded.txt"}, follow_redirects=False)
    assert resp.status_code == 404



"""Unit tests for user actor interaction tracking."""

from __future__ import annotations

from types import SimpleNamespace
from unittest.mock import MagicMock

from bloom_lims.domain.base import BloomObj


def _build_bloom_obj_stub() -> BloomObj:
    bo = BloomObj.__new__(BloomObj)
    bo.logger = MagicMock()
    bo.session = MagicMock()
    bo._actor_user_id = None
    bo._actor_email = None
    return bo


def test_track_user_interaction_creates_lineage_from_actor():
    bo = _build_bloom_obj_stub()
    bo._upsert_user_actor = MagicMock(return_value=SimpleNamespace(euid="AX-1"))
    bo.get_by_euid = MagicMock(
        return_value=SimpleNamespace(euid="CX-35", is_deleted=False)
    )
    bo.create_generic_instance_lineage_by_euids = MagicMock()

    actor_euid = BloomObj.track_user_interaction(
        bo,
        "CX-35",
        relationship_type="user_created",
        action_ds={"curr_user": {"email": "user@lsmc.com", "cognito_sub": "sub-1"}},
    )

    assert actor_euid == "AX-1"
    bo._upsert_user_actor.assert_called_once_with(user_id="sub-1", email="user@lsmc.com")
    bo.create_generic_instance_lineage_by_euids.assert_called_once_with(
        "AX-1",
        "CX-35",
        relationship_type="user_created",
    )
    bo.session.commit.assert_called_once()


def test_track_user_interaction_skips_when_target_not_found():
    bo = _build_bloom_obj_stub()
    bo._upsert_user_actor = MagicMock(return_value=SimpleNamespace(euid="AX-1"))
    bo.get_by_euid = MagicMock(return_value=None)
    bo.create_generic_instance_lineage_by_euids = MagicMock()

    actor_euid = BloomObj.track_user_interaction(bo, "MISSING-EUID")

    assert actor_euid is None
    bo.create_generic_instance_lineage_by_euids.assert_not_called()
    bo.session.commit.assert_not_called()


def test_do_action_base_tracks_user_action_lineage(monkeypatch):
    bo = _build_bloom_obj_stub()
    action_name = "action/core/set_object_status/1.0"
    now_dt = "2026-03-03 12:34:56"
    action_ds = {"deactivate_actions_when_executed": [], "curr_user": "user@lsmc.com"}
    bobj = SimpleNamespace(
        json_addl={
            "action_groups": {
                "core": {
                    "actions": {
                        action_name: {
                            "action_executed": "0",
                            "max_executions": "-1",
                            "executed_datetime": [],
                            "action_user": [],
                            "action_enabled": "1",
                        }
                    }
                }
            }
        }
    )
    bo.get_by_euid = MagicMock(return_value=bobj)
    bo.track_user_interaction = MagicMock()
    monkeypatch.setattr("bloom_lims.domain.base.flag_modified", lambda *_args, **_kwargs: None)

    result = BloomObj._do_action_base(bo, "CX-35", action_name, "core", action_ds, now_dt)

    assert result is bobj
    action_state = bobj.json_addl["action_groups"]["core"]["actions"][action_name]
    assert action_state["action_executed"] == "1"
    assert action_state["executed_datetime"] == [now_dt]
    assert action_state["action_user"] == ["user@lsmc.com"]
    bo.session.commit.assert_called_once()
    bo.track_user_interaction.assert_called_once_with(
        "CX-35",
        relationship_type=f"user_action:{action_name}",
        action_ds=action_ds,
    )
